# Module:   help
# Date:     28th November 2013
# Author:   James Mills, j dot mills at griffith dot edu dot au

"""Help Tasks"""


from __future__ import print_function


from fabric import state
from fabric.api import task
from fabric.tasks import Task
from fabric.task_utils import crawl


@task(default=True)
def help(name=None):
    """Display help for a given task

    Options:
        name    - The task to display help on.

    To display a list of available tasks type:

        $ fab -l

    To display help on a specific task type:

        $ fab help:<name>
    """

    if name is None:
        name = "help"

    task = crawl(name, state.commands)
    if isinstance(task, Task):
        doc = getattr(task, "__doc__", None)
        if doc is not None:
            print("Help on {0:s}:".format(name))
            print()
            print(doc)
        else:
            print("No help available for {0;s}".format(name))
    else:
        print("No such task {0:s}".format(name))
        print("For a list of tasks type: fab -l")



#!/usr/bin/env python

from circuits.web import Controller

from .helpers import urlopen


class Root(Controller):

    def index(self):
        return "Hello World!"

    def foo(self):
        self.response.headers["Content-Type"] = "text/plain"
        return "Hello World!"

    def empty(self):
        return ""


def test_default(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    content_type = f.headers["Content-Type"]
    assert content_type == "text/html; charset=utf-8"


def test_explicit(webapp):
    f = urlopen("{0:s}/foo".format(webapp.server.http.base))
    s = f.read()
    assert s == b"Hello World!"

    content_type = f.headers["Content-Type"]
    assert content_type == "text/plain"


def test_static(webapp):
    f = urlopen("{0:s}/static/test.css".format(webapp.server.http.base))
    s = f.read()
    assert s == b"body { }\n"

    content_type = f.headers["Content-Type"]
    assert content_type == "text/css"


def test_empty(webapp):
    f = urlopen("{0:s}/empty".format(webapp.server.http.base))
    s = f.read()
    assert s == b""

    content_length = f.headers["Content-Length"]
    assert int(content_length) == 0



#!/usr/bin/env python

from socket import gaierror


def test_client_bind_int():
    from circuits.net.sockets import Client

    class TestClient(Client):

        def _create_socket(self):
            return None

    client = TestClient(1234)

    assert client._bind[1] == 1234


def test_client_bind_int_gaierror(monkeypatch):
    from circuits.net import sockets

    def broken_gethostname():
        raise gaierror()

    monkeypatch.setattr(sockets, "gethostname", broken_gethostname)

    class TestClient(sockets.Client):

        def _create_socket(self):
            return None

    client = TestClient(1234)

    assert client._bind == ("0.0.0.0", 1234)


def test_client_bind_str():
    from circuits.net.sockets import Client

    class TestClient(Client):

        def _create_socket(self):
            return None

    client = TestClient("0.0.0.0:1234")

    assert client._bind == ("0.0.0.0", 1234)



# Module:   line
# Date:     04th February 2010
# Author:   James Mills <prologic@shortcircuit.net.au>

"""Line Protocol

This module implements the basic Line protocol.

This module can be used in both server and client implementations.
"""

import re

from circuits.core import handler, Event, BaseComponent

from circuits.six import b


LINESEP = re.compile(b("\r?\n"))


def splitLines(s, buffer):
    """splitLines(s, buffer) -> lines, buffer

    Append s to buffer and find any new lines of text in the
    string splitting at the standard IRC delimiter CRLF. Any
    new lines found, return them as a list and the remaining
    buffer for further processing.
    """

    lines = LINESEP.split(buffer + s)
    return lines[:-1], lines[-1]


class line(Event):
    """line Event"""


class Line(BaseComponent):
    """Line Protocol

    Implements the Line Protocol.

    Incoming data is split into lines with a splitter function. For each
    line of data processed a Line Event is created. Any unfinished lines
    are appended into an internal buffer.

    A custom line splitter function can be passed to customize how data
    is split into lines. This function must accept two arguments, the data
    to process and any left over data from a previous invocation of the
    splitter function. The function must also return a tuple of two items,
    a list of lines and any left over data.

    :param splitter: a line splitter function
    :type  splitter: function

    This Component operates in two modes. In normal operation it's expected
    to be used in conjunction with components that expose a Read Event on
    a "read" channel with only one argument (data). Some builtin components
    that expose such events are:
    - circuits.net.sockets.TCPClient
    - circuits.io.File

    The second mode of operation works with circuits.net.sockets.Server
    components such as TCPServer, UNIXServer, etc. It's expected that
    two arguments exist in the Read Event, sock and data. The following
    two arguments can be passed to affect how unfinished data is stored
    and retrieved for such components:

    :param getBuffer: function to retrieve the buffer for a client sock
    :type getBuffer: function

    This function must accept one argument (sock,) the client socket
    whoose buffer is to be retrieved.

    :param updateBuffer: function to update the buffer for a client sock
    :type updateBuffer: function

    This function must accept two arguments (sock, buffer,) the client
    socket and the left over buffer to be updated.

    @note: This Component must be used in conjunction with a Component that
           exposes Read events on a "read" Channel.
    """

    def __init__(self, *args, **kwargs):
        "initializes x; see x.__class__.__doc__ for signature"

        super(Line, self).__init__(*args, **kwargs)

        self.encoding = kwargs.get('encoding', 'utf-8')

        # Used for Servers
        self.getBuffer = kwargs.get("getBuffer")
        self.updateBuffer = kwargs.get("updateBuffer")

        self.splitter = kwargs.get("splitter", splitLines)

        self.buffer = b""

    @handler("read")
    def _on_read(self, *args):
        if len(args) == 1:
            # Client read
            data, = args
            lines, self.buffer = self.splitter(data, self.buffer)
            [self.fire(line(x)) for x in lines]
        else:
            # Server read
            sock, data = args
            lines, buffer = self.splitter(data, self.getBuffer(sock))
            self.updateBuffer(sock, buffer)
            [self.fire(line(sock, x)) for x in lines]



#!/usr/bin/env python

from circuits.core.manager import Manager
from circuits.core.handlers import handler
from circuits.core.components import BaseComponent

from circuits.web import BaseServer, Controller
from circuits.web.dispatchers.dispatcher import Dispatcher

from .helpers import urlopen, urljoin


class PrefixingDispatcher(BaseComponent):
    """Forward to another Dispatcher based on the channel."""

    def __init__(self, channel):
        super(PrefixingDispatcher, self).__init__(channel=channel)

    @handler("request", priority=1.0)
    def _on_request(self, event, request, response):
        path = request.path.strip("/")

        path = urljoin("/%s/" % self.channel, path)
        request.path = path


class DummyRoot(Controller):

    channel = "/"

    def index(self):
        return "Not used"


class Root1(Controller):

    channel = "/site1"

    def index(self):
        return "Hello from site 1!"


class Root2(Controller):

    channel = "/site2"

    def index(self):
        return "Hello from site 2!"


def test_disps():

    manager = Manager()

    server1 = BaseServer(0, channel="site1")
    server1.register(manager)
    PrefixingDispatcher(channel="site1").register(server1)
    Dispatcher(channel="site1").register(server1)
    Root1().register(manager)

    server2 = BaseServer(("localhost", 0), channel="site2")
    server2.register(manager)
    PrefixingDispatcher(channel="site2").register(server2)
    Dispatcher(channel="site2").register(server2)
    Root2().register(manager)

    DummyRoot().register(manager)
    manager.start()

    f = urlopen(server1.http.base, timeout=3)
    s = f.read()
    assert s == b"Hello from site 1!"

    f = urlopen(server2.http.base, timeout=3)
    s = f.read()
    assert s == b"Hello from site 2!"



#!/usr/bin/env python

import pytest
if pytest.PLATFORM == "win32":
    pytest.skip("Unsupported Platform")

from circuits.io import Process, write


def test(manager, watcher):
    p = Process(["echo", "Hello World!"]).register(manager)
    assert watcher.wait("registered")

    p.start()
    assert watcher.wait("started", p.channel)

    assert watcher.wait("stopped", p.channel)

    s = p.stdout.getvalue()
    assert s == b"Hello World!\n"


def test2(manager, watcher, tmpdir):
    foo = tmpdir.ensure("foo.txt")

    p = Process(["cat - > {0:s}".format(str(foo))], shell=True).register(manager)
    assert watcher.wait("registered")

    p.start()
    assert watcher.wait("started", p.channel)

    p.fire(write("Hello World!"), p._stdin)
    assert watcher.wait("write", p._stdin)

    p.stop()

    assert watcher.wait("eof", p._stdout.channel)

    with foo.open("r") as f:
        assert f.read() == "Hello World!"



#!/usr/bin/env python

from circuits.web import Controller

from .helpers import urlopen


class Root(Controller):

    def index(self):
        def response():
            yield "Hello "
            yield "World!"
        return response()


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"



# Module:   test_timers
# Date:     10th February 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au
import time

"""Timers Tests"""

import pytest

from datetime import datetime, timedelta

from circuits import Event, Component, Timer


def pytest_funcarg__app(request):
    return request.cached_setup(
        setup=lambda: setupapp(request),
        teardown=lambda app: teardownapp(app),
        scope="module"
    )


def setupapp(request):
    app = App()
    app.start()
    return app


def teardownapp(app):
    app.stop()


class test(Event):
    """test Event"""


class App(Component):

    def __init__(self):
        super(App, self).__init__()
        self.flag = False
        self.count = 0
        self.timestamps = []

    def reset(self):
        self.timestamps = []
        self.flag = False
        self.count = 0

    def test(self):
        self.timestamps.append(time.time())
        self.count += 1
        self.flag = True


def test_timer(app):
    timer = Timer(0.1, test(), "timer")
    timer.register(app)
    assert pytest.wait_for(app, "flag")
    app.reset()


def test_persistentTimer(app):
    app.timestamps.append(time.time())
    timer = Timer(0.2, test(), "timer", persist=True)
    timer.register(app)

    wait_res = pytest.wait_for(app, "count", 2)
    assert app.count >= 2
    assert wait_res
    delta = app.timestamps[1] - app.timestamps[0]
    # Should be 0.1, but varies depending on timer precision and load
    assert delta >= 0.08 and delta < 0.5
    delta = app.timestamps[2] - app.timestamps[1]
    assert delta >= 0.08 and delta < 0.5
    app.reset()

    timer.unregister()


def test_datetime(app):
    now = datetime.now()
    d = now + timedelta(seconds=0.1)
    timer = Timer(d, test(), "timer")
    timer.register(app)
    assert pytest.wait_for(app, "flag")
    app.reset()



#!/usr/bin/env python

from json import dumps

from circuits import handler, Component
from circuits.web import Server, Controller, Logger


class JSONSerializer(Component):

    channel = "web"

    @handler("response", priority=1.0)  # 1 higher than the default response handler
    def serialize_response_body(self, response):
        response.headers["Content-Type"] = "application/json"
        response.body = dumps(response.body)


class Root(Controller):

    def index(self):
        return {"message": "Hello World!"}

app = Server(("0.0.0.0", 9000))
JSONSerializer().register(app)
Logger().register(app)
Root().register(app)
app.run()



try:
    from urllib.error import HTTPError, URLError
    from urllib.parse import quote, urlencode, urljoin
    from urllib.request import HTTPBasicAuthHandler, HTTPCookieProcessor
    from urllib.request import urlopen, build_opener, install_opener
    from urllib.request import HTTPDigestAuthHandler, Request
except ImportError:
    from urlparse import urljoin
    from urllib import quote, urlencode
    from urllib2 import HTTPError, URLError, HTTPDigestAuthHandler
    from urllib2 import HTTPBasicAuthHandler, HTTPCookieProcessor
    from urllib2 import urlopen, build_opener, install_opener, Request

try:
    from http.cookiejar import CookieJar
except ImportError:
    from cookielib import CookieJar

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

# pylama:skip=1



#!/usr/bin/env python

from uuid import uuid4 as uuid

from circuits import Component
from circuits.net.events import close, connect, write
from circuits.net.sockets import TCPClient, TCPServer


class Client(Component):

    channel = "client"

    def init(self, sock, host, port, channel=channel):
        self.sock = sock
        self.host = host
        self.port = port

        TCPClient(channel=self.channel).register(self)

    def ready(self, *args):
        self.fire(connect(self.host, self.port))

    def disconnect(self, *args):
        self.fire(close(self.sock), self.parent.channel)

    def read(self, data):
        self.fire(write(self.sock, data), self.parent.channel)


class Proxy(Component):

    channel = "server"

    def init(self, bind, host, port):
        self.bind = bind
        self.host = host
        self.port = port

        self.clients = dict()

        TCPServer(self.bind).register(self)

    def connect(self, sock, host, port):
        channel = uuid()

        client = Client(
            sock, self.host, self.port, channel=channel
        ).register(self)

        self.clients[sock] = client

    def disconnect(self, sock):
        client = self.clients.get(sock)
        if client is not None:
            client.unregister()
            del self.clients[sock]

    def read(self, sock, data):
        client = self.clients[sock]
        self.fire(write(data), client.channel)


app = Proxy(("0.0.0.0", 3333), "127.0.0.1", 22)

from circuits import Debugger
Debugger().register(app)

app.run()



#!/usr/bin/env python


import pytest
from pytest import fixture


from circuits import handler, Event, Component

from circuits.net.events import read, write

from circuits.protocols.irc import IRC
from circuits.protocols.irc import strip, joinprefix, parseprefix

from circuits.protocols.irc import (
    PASS, USER, NICK, PONG, QUIT,
    JOIN, PART, PRIVMSG, NOTICE, AWAY,
    KICK, TOPIC, MODE, INVITE, NAMES, WHOIS
)

from circuits.six import u


class App(Component):

    def init(self):
        IRC().register(self)

        self.data = []
        self.events = []

    @handler(False)
    def reset(self):
        self.data = []
        self.events = []

    @handler()
    def _on_event(self, event, *args, **kwargs):
        self.events.append(event)

    def request(self, message):
        self.fire(write(bytes(message)))

    def write(self, data):
        self.data.append(data)


@fixture(scope="function")
def app(request):
    app = App()

    while app:
        app.flush()

    return app


def test_strip():
    s = ":\x01\x02test\x02\x01"
    s = strip(s)
    assert s == "\x01\x02test\x02\x01"

    s = ":\x01\x02test\x02\x01"
    s = strip(s, color=True)
    assert s == "test"


def test_joinprefix():
    nick, ident, host = "test", "foo", "localhost"
    s = joinprefix(nick, ident, host)
    assert s == "test!foo@localhost"


def test_parseprefix():
    s = "test!foo@localhost"
    nick, ident, host = parseprefix(s)
    assert nick == "test"
    assert ident == "foo"
    assert host == "localhost"

    s = "test"
    nick, ident, host = parseprefix(s)
    assert nick == "test"
    assert ident is None
    assert host is None


@pytest.mark.parametrize("event,data", [
    (PASS("secret"), b"PASS secret\r\n"),
    (
        USER("foo", "localhost", "localhost", "Test Client"),
        b"USER foo localhost localhost :Test Client\r\n"
    ),
    (NICK("test"), b"NICK test\r\n"),
    (PONG("localhost"), b"PONG :localhost\r\n"),
    (QUIT(), b"QUIT Leaving\r\n"),
    (QUIT("Test"), b"QUIT Test\r\n"),
    (QUIT("Test Message"), b"QUIT :Test Message\r\n"),
    (JOIN("#test"), b"JOIN #test\r\n"),
    (JOIN("#test", "secret"), b"JOIN #test secret\r\n"),
    (PART("#test"), b"PART #test\r\n"),
    (PRIVMSG("test", "Hello"), b"PRIVMSG test Hello\r\n"),
    (PRIVMSG("test", "Hello World"), b"PRIVMSG test :Hello World\r\n"),
    (NOTICE("test", "Hello"), b"NOTICE test Hello\r\n"),
    (NOTICE("test", "Hello World"), b"NOTICE test :Hello World\r\n"),
    (KICK("#test", "test"), b"KICK #test test :\r\n"),
    (KICK("#test", "test", "Bye"), b"KICK #test test Bye\r\n"),
    (KICK("#test", "test", "Good Bye!"), b"KICK #test test :Good Bye!\r\n"),
    (TOPIC("#test", "Hello World!"), b"TOPIC #test :Hello World!\r\n"),
    (MODE("+i"), b"MODE +i\r\n"),
    (MODE("#test", "+o", "test"), b"MODE #test +o test\r\n"),
    (INVITE("test", "#test"), b"INVITE test #test\r\n"),
    (NAMES(), b"NAMES\r\n"),
    (NAMES("#test"), b"NAMES #test\r\n"),
    (AWAY("I am away."), b"AWAY :I am away.\r\n"),
    (WHOIS("somenick"), b"WHOIS :somenick\r\n"),
])
def test_commands(event, data):
    message = event.args[0]
    return bytes(message) == data


@pytest.mark.parametrize("data,event", [
    (
        b":localhost NOTICE * :*** Looking up your hostname...\r\n",
        Event.create(
            "notice", (u("localhost"), None, None), u("*"),
            u("*** Looking up your hostname..."),
        )
    ),
])
def test_responses(app, data, event):
    app.reset()
    app.fire(read(data))
    while app:
        app.flush()

    e = app.events[-1]

    assert event.name == e.name
    assert event.args == e.args
    assert event.kwargs == e.kwargs



#!/usr/bin/env python

import pytest
if pytest.PLATFORM == "win32":
    pytest.skip("Unsupported Platform")

from circuits import Manager
from circuits.net.sockets import Pipe
from circuits.core.pollers import Select
from circuits.net.events import close, write

from .client import Client


def pytest_generate_tests(metafunc):
    metafunc.addcall(funcargs={"Poller": Select})


def test_pipe(Poller):
    m = Manager() + Poller()

    a, b = Pipe("a", "b")
    a.register(m)
    b.register(m)

    a = Client(channel=a.channel).register(m)
    b = Client(channel=b.channel).register(m)

    m.start()

    try:
        assert pytest.wait_for(a, "ready")
        assert pytest.wait_for(b, "ready")

        a.fire(write(b"foo"))
        assert pytest.wait_for(b, "data", b"foo")

        b.fire(write(b"foo"))
        assert pytest.wait_for(a, "data", b"foo")

        a.fire(close())
        assert pytest.wait_for(a, "disconnected")

        b.fire(close())
        assert pytest.wait_for(b, "disconnected")
    finally:
        m.stop()



# Module:   io
# Date:     4th August 2004
# Author:   James Mills <prologic@shortcircuit.net.au>

"""I/O Support

This package contains various I/O Components. Provided are a generic File
Component, StdIn, StdOut and StdErr components. Instances of StdIn, StdOut
and StdErr are also created by importing this package.
"""

import sys

from .file import File
from .process import Process
from .events import close, open, seek, write

try:
    from .notify import Notify
except:
    pass

try:
    from .serial import Serial
except:
    pass

try:
    stdin = File(sys.stdin, channel="stdin")
    stdout = File(sys.stdout, channel="stdout")
    stderr = File(sys.stderr, channel="stderr")
except:
    pass

# flake8: noqa



# Module:   jsonrpc
# Date:     13th September 2007
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""JSON RPC

This module implements a JSON RPC dispatcher that translates incoming
RPC calls over JSON into RPC events.
"""

from circuits.tools import tryimport

json = tryimport(("json", "simplejson"))

from circuits.six import binary_type
from circuits import handler, Event, BaseComponent


class rpc(Event):
    """RPC Event"""


class JSONRPC(BaseComponent):

    channel = "web"

    def __init__(self, path=None, encoding="utf-8", rpc_channel="*"):
        super(JSONRPC, self).__init__()

        if json is None:
            raise RuntimeError("No json support available")

        self.path = path
        self.encoding = encoding
        self.rpc_channel = rpc_channel

    @handler("request", priority=0.2)
    def _on_request(self, event, req, res):
        if self.path is not None and self.path != req.path.rstrip("/"):
            return

        res.headers["Content-Type"] = "application/json"

        try:
            data = req.body.read().decode(self.encoding)
            o = json.loads(data)
            id, method, params = o["id"], o["method"], o["params"]
            if isinstance(params, dict):
                params = dict([(str(k), v) for k, v in params.iteritems()])

            if "." in method:
                channel, name = method.split(".", 1)
            else:
                channel, name = self.rpc_channel, method

            name = str(name) if not isinstance(name, binary_type) else name

            if isinstance(params, dict):
                value = yield self.call(rpc.create(name, **params), channel)
            else:
                value = yield self.call(rpc.create(name, *params), channel)

            yield self._response(id, value.value)
        except Exception as e:
            yield self._error(-1, 100, "%s: %s" % (e.__class__.__name__, e))
        finally:
            event.stop()

    def _response(self, id, result):
        data = {
            "id": id,
            "version": "1.1",
            "result": result,
            "error": None
        }
        return json.dumps(data).encode(self.encoding)

    def _error(self, id, code, message):
        data = {
            "id": id,
            "version": "1.1",
            "error": {
                "name": "JSONRPCError",
                "code": code,
                "message": message
            }
        }
        return json.dumps(data).encode(self.encoding)



# Module:   exceptions
# Date:     10th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Exceptions

This module implements a set of standard HTTP Errors as Python Exceptions.

Note: This code is mostly borrowed from werkzeug and adapted for circuits.web
"""

from inspect import isclass

from .constants import HTTP_STATUS_CODES


class HTTPException(Exception):
    """
    Baseclass for all HTTP exceptions.  This exception can be called by WSGI
    applications to render a default error page or you can catch the subclasses
    of it independently and render nicer error messages.
    """

    code = None
    traceback = True
    description = None

    def __init__(self, description=None, traceback=None):
        super(HTTPException, self).__init__("%d %s" % (self.code, self.name))
        if description is not None:
            self.description = description
        if traceback is not None:
            self.traceback = traceback

    @property
    def name(self):
        """The status name."""
        return HTTP_STATUS_CODES.get(self.code, '')

    def __repr__(self):
        return '<%s \'%s\'>' % (self.__class__.__name__, self)


class BadRequest(HTTPException):
    """*400* `Bad Request`

    Raise if the browser sends something to the application the application
    or server cannot handle.
    """
    code = 400
    description = (
        '<p>The browser (or proxy) sent a request that this server could '
        'not understand.</p>'
    )


class UnicodeError(HTTPException):
    """
    raised by the request functions if they were unable to decode the
    incoming data properly.
    """


class Unauthorized(HTTPException):
    """*401* `Unauthorized`

    Raise if the user is not authorized.  Also used if you want to use HTTP
    basic auth.
    """
    code = 401
    description = (
        '<p>The server could not verify that you are authorized to access '
        'the URL requested.  You either supplied the wrong credentials (e.g. '
        'a bad password), or your browser doesn\'t understand how to supply '
        'the credentials required.</p><p>In case you are allowed to request '
        'the document, please check your user-id and password and try '
        'again.</p>'
    )


class Forbidden(HTTPException):
    """*403* `Forbidden`

    Raise if the user doesn't have the permission for the requested resource
    but was authenticated.
    """
    code = 403
    description = (
        '<p>You don\'t have the permission to access the requested resource. '
        'It is either read-protected or not readable by the server.</p>'
    )


class NotFound(HTTPException):
    """*404* `Not Found`

    Raise if a resource does not exist and never existed.
    """
    code = 404
    description = (
        '<p>The requested URL was not found on the server.</p>'
        '<p>If you entered the URL manually please check your spelling and '
        'try again.</p>'
    )


class MethodNotAllowed(HTTPException):
    """*405* `Method Not Allowed`

    Raise if the server used a method the resource does not handle.  For
    example `POST` if the resource is view only.  Especially useful for REST.

    The first argument for this exception should be a list of allowed methods.
    Strictly speaking the response would be invalid if you don't provide valid
    methods in the header which you can do with that list.
    """
    code = 405

    def __init__(self, method, description=None):
        HTTPException.__init__(self, description)
        if description is None:
            self.description = (
                '<p>The method %s is not allowed '
                'for the requested URL.</p>'
            ) % method


class NotAcceptable(HTTPException):
    """*406* `Not Acceptable`

    Raise if the server can't return any content conforming to the
    `Accept` headers of the client.
    """
    code = 406

    description = (
        '<p>The resource identified by the request is only capable of '
        'generating response entities which have content characteristics '
        'not acceptable according to the accept headers sent in the '
        'request.</p>'
    )


class RequestTimeout(HTTPException):
    """*408* `Request Timeout`

    Raise to signalize a timeout.
    """
    code = 408
    description = (
        '<p>The server closed the network connection because the browser '
        'didn\'t finish the request within the specified time.</p>'
    )


class Gone(HTTPException):
    """*410* `Gone`

    Raise if a resource existed previously and went away without new location.
    """
    code = 410
    description = (
        '<p>The requested URL is no longer available on this server and '
        'there is no forwarding address.</p><p>If you followed a link '
        'from a foreign page, please contact the author of this page.'
    )


class LengthRequired(HTTPException):
    """*411* `Length Required`

    Raise if the browser submitted data but no ``Content-Length`` header which
    is required for the kind of processing the server does.
    """
    code = 411
    description = (
        '<p>A request with this method requires a valid <code>Content-'
        'Length</code> header.</p>'
    )


class PreconditionFailed(HTTPException):
    """*412* `Precondition Failed`

    Status code used in combination with ``If-Match``, ``If-None-Match``, or
    ``If-Unmodified-Since``.
    """
    code = 412
    description = (
        '<p>The precondition on the request for the URL failed positive '
        'evaluation.</p>'
    )


class RequestEntityTooLarge(HTTPException):
    """*413* `Request Entity Too Large`

    The status code one should return if the data submitted exceeded a given
    limit.
    """
    code = 413
    description = (
        '<p>The data value transmitted exceeds the capacity limit.</p>'
    )


class RequestURITooLarge(HTTPException):
    """*414* `Request URI Too Large`

    Like *413* but for too long URLs.
    """
    code = 414
    description = (
        '<p>The length of the requested URL exceeds the capacity limit '
        'for this server.  The request cannot be processed.</p>'
    )


class UnsupportedMediaType(HTTPException):
    """*415* `Unsupported Media Type`

    The status code returned if the server is unable to handle the media type
    the client transmitted.
    """
    code = 415
    description = (
        '<p>The server does not support the media type transmitted in '
        'the request.</p>'
    )


class RangeUnsatisfiable(HTTPException):
    """*416* `Range Unsatisfiable`

    The status code returned if the server is unable to satisfy the request range
    """

    code = 416
    description = (
        '<p>The server cannot satisfy the request range(s).</p>'
    )


class InternalServerError(HTTPException):
    """*500* `Internal Server Error`

    Raise if an internal server error occurred.  This is a good fallback if an
    unknown error occurred in the dispatcher.
    """
    code = 500
    description = (
        '<p>The server encountered an internal error and was unable to '
        'complete your request.  Either the server is overloaded or there '
        'is an error in the application.</p>'
    )


class NotImplemented(HTTPException):
    """*501* `Not Implemented`

    Raise if the application does not support the action requested by the
    browser.
    """
    code = 501
    description = (
        '<p>The server does not support the action requested by the '
        'browser.</p>'
    )


class BadGateway(HTTPException):
    """*502* `Bad Gateway`

    If you do proxying in your application you should return this status code
    if you received an invalid response from the upstream server it accessed
    in attempting to fulfill the request.
    """
    code = 502
    description = (
        '<p>The proxy server received an invalid response from an upstream '
        'server.</p>'
    )


class ServiceUnavailable(HTTPException):
    """*503* `Service Unavailable`

    Status code you should return if a service is temporarily unavailable.
    """
    code = 503
    description = (
        '<p>The server is temporarily unable to service your request due to '
        'maintenance downtime or capacity problems.  Please try again '
        'later.</p>'
    )


class Redirect(HTTPException):

    code = 303

    def __init__(self, urls, status=None):
        super(Redirect, self).__init__()

        if isinstance(urls, str):
            self.urls = [urls]
        else:
            self.urls = urls

        self.status = status

__all__ = [
    x[0] for x in list(globals().items())
    if isclass(x[1]) and issubclass(x[1], HTTPException)
]



#!/usr/bin/env python

from circuits.web import Controller
from circuits.web.exceptions import Forbidden, NotFound, Redirect

from .helpers import urlopen, HTTPError


class Root(Controller):

    def index(self):
        return "Hello World!"

    def test_redirect(self):
        raise Redirect("/")

    def test_forbidden(self):
        raise Forbidden()

    def test_notfound(self):
        raise NotFound()

    def test_contenttype(self):
        self.response.headers["Content-Type"] = "application/json"
        raise Exception()


def test_redirect(webapp):
    f = urlopen("%s/test_redirect" % webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"


def test_forbidden(webapp):
    try:
        urlopen("%s/test_forbidden" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 403
        assert e.msg == "Forbidden"
    else:
        assert False


def test_notfound(webapp):
    try:
        urlopen("%s/test_notfound" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == "Not Found"
    else:
        assert False


def test_contenttype(webapp):
    try:
        f = urlopen("%s/test_contenttype" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 500
        assert e.msg == "Internal Server Error"
        assert e.headers.get("Content-Type") == "text/html"
    else:
        assert False



"""HTML macros

Macros for generating snippets of HTML.
"""

import genshi
import pygments
import pygments.util
import pygments.lexers
import pygments.formatters
from genshi import builder
from genshi.filters import HTMLSanitizer

sanitizer = HTMLSanitizer()


def pre(macro, environ, *args, **kwargs):
    """Return the raw text of body, rendered in a <pre> block.

    **Arguments:** //None//

    **Example:**
    {{{
    <<pre>>
    def hello():
        print "Hello World!"

    hello()
    <</pre>>
    }}}
    """

    if macro.body is None:
        return None

    return builder.tag.pre(macro.body)


def code(macro, environ, *args, **kwargs):
    """Render syntax highlighted code"""

    if not macro.body:
        return None

    lang = kwargs.get("lang", None)

    if lang is not None:
        if not macro.isblock:
            return None
        try:
            lexer = pygments.lexers.get_lexer_by_name(lang, stripall=True)
        except pygments.util.ClassNotFound:
            return None
    else:
        lexer = None

    if lexer:
        text = pygments.highlight(
            macro.body, lexer,
            pygments.formatters.HtmlFormatter()
        )
        output = genshi.core.Markup(text)
    elif macro.isblock:
        output = genshi.builder.tag.pre(macro.body)
    else:
        output = genshi.builder.tag.code(
            macro.body,
            style="white-space:pre-wrap", class_="highlight"
        )

    return output


def source(macro, environ, *args, **kwargs):
    """Return the parsed text of body, rendered in a <pre> block."""

    if macro.body is None:
        return None

    return builder.tag.pre(environ["parser"].render(
        macro.body, environ=environ).decode("utf-8"))


def div(macro, environ, cls=None, float=None, id=None, style=None,
        *args, **kwargs):

    if macro.body is None:
        return None

    if float and float in ("left", "right"):
        style = "float: %s; %s" % (float, style)

    if style:
        style = ";".join(sanitizer.sanitize_css(style))

    if macro.isblock:
        context = "block"
    else:
        context = "inline"

    contents = environ["parser"].generate(
        macro.body, environ=environ, context=context
    )

    return builder.tag.div(contents, id=id, class_=cls, style=style)


def span(macro, environ, class_=None, id=None, style=None, *args, **kwargs):
    """..."""

    if macro.body is None:
        return None

    if style:
        style = ';'.join(sanitizer.sanitize_css(style))

    contents = environ['parser'].generate(
        macro.body, environ=environ, context='inline'
    )

    return builder.tag.span(contents, id=id, class_=class_, style=style)



#!/usr/bin/env python

from circuits.web import Server, Controller, Sessions


class Root(Controller):

    def index(self, name="World"):
        if "name" in self.session:
            name = self.session["name"]
        self.session["name"] = name
        return "Hello %s!" % name

app = Server(("0.0.0.0", 8000))
Sessions().register(app)
Root().register(app)
app.run()



# Package:  websockets
# Date:     26th March 2013
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""circuits.web websockets"""

from .client import WebSocketClient
from .dispatcher import WebSocketsDispatcher

# flake8: noqa



#!/usr/bin/env python

import pytest

from os import path
from socket import gaierror

from circuits.web import Controller
from circuits import handler, Component
from circuits.web import BaseServer, Server

from .helpers import urlopen, URLError

CERTFILE = path.join(path.dirname(__file__), "cert.pem")


class BaseRoot(Component):

    channel = "web"

    def request(self, request, response):
        return "Hello World!"


class Root(Controller):

    def index(self):
        return "Hello World!"


class MakeQuiet(Component):

    @handler("ready", channel="*", priority=1.0)
    def _on_ready(self, event, *args):
        event.stop()


def test_baseserver(manager, watcher):
    server = BaseServer(0).register(manager)
    MakeQuiet().register(server)
    watcher.wait("ready")

    BaseRoot().register(server)
    watcher.wait("registered")

    try:
        f = urlopen(server.http.base)
    except URLError as e:
        if type(e[0]) is gaierror:
            f = urlopen("http://127.0.0.1:9000")
        else:
            raise

    s = f.read()
    assert s == b"Hello World!"

    server.unregister()
    watcher.wait("unregistered")


def test_server(manager, watcher):
    server = Server(0).register(manager)
    MakeQuiet().register(server)
    watcher.wait("ready")

    Root().register(server)

    try:
        f = urlopen(server.http.base)
    except URLError as e:
        if type(e[0]) is gaierror:
            f = urlopen("http://127.0.0.1:9000")
        else:
            raise

    s = f.read()
    assert s == b"Hello World!"

    server.unregister()
    watcher.wait("unregistered")


def test_secure_server(manager, watcher):
    pytest.importorskip("ssl")

    server = Server(0, secure=True, certfile=CERTFILE).register(manager)
    MakeQuiet().register(server)
    watcher.wait("ready")

    Root().register(server)

    try:
        f = urlopen(server.http.base)
    except URLError as e:
        if type(e[0]) is gaierror:
            f = urlopen("http://127.0.0.1:9000")
        else:
            raise

    s = f.read()
    assert s == b"Hello World!"

    server.unregister()
    watcher.wait("unregistered")


def test_unixserver(manager, watcher, tmpdir):
    if pytest.PLATFORM == "win32":
        pytest.skip("Unsupported Platform")

    sockpath = tmpdir.ensure("test.sock")
    socket = str(sockpath)

    server = Server(socket).register(manager)
    MakeQuiet().register(server)
    watcher.wait("ready")

    Root().register(server)

    assert path.basename(server.host) == "test.sock"

    server.unregister()
    watcher.wait("unregistered")


def test_multi_servers(manager, watcher):
    pytest.importorskip("ssl")

    insecure_server = Server(0, channel="insecure")
    secure_server = Server(
        0,
        channel="secure", secure=True, certfile=CERTFILE
    )

    server = (insecure_server + secure_server).register(manager)
    MakeQuiet().register(server)
    watcher.wait("ready")

    Root().register(server)

    f = urlopen(insecure_server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    f = urlopen(secure_server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    server.unregister()
    watcher.wait("unregistered")



#!/usr/bin/env python

from circuits import Component, Event


class hello(Event):
    """hello Event"""


class foo(Event):
    """foo Event"""


class bar(Event):
    """bar Event"""


class App(Component):

    def foo(self):
        print("Foo!")

    def bar(self):
        print("Bar!")

    def hello(self):
        self.fire(foo())
        self.fire(bar())
        print("Hello World!")

    def started(self, component):
        self.fire(hello())
        self.stop()

App().run()



#!/usr/bin/env python


from pytest import fixture


from os import path
from io import BytesIO


from circuits.web import Controller
from circuits.web.tools import gzip
from circuits import handler, Component


from .conftest import DOCROOT
from .helpers import build_opener, Request


class Gzip(Component):

    channel = "web"

    @handler("response", priority=1.0)
    def _on_response(self, event, *args, **kwargs):
        event[0] = gzip(event[0])


class Root(Controller):

    def index(self):
        return "Hello World!"


@fixture(scope="module")
def gziptool(request, webapp):
    gziptool = Gzip().register(webapp)

    def finalizer():
        gziptool.unregister()

    request.addfinalizer(finalizer)

    return gziptool


def decompress(body):
    import gzip

    zbuf = BytesIO()
    zbuf.write(body)
    zbuf.seek(0)
    zfile = gzip.GzipFile(mode='rb', fileobj=zbuf)
    data = zfile.read()
    zfile.close()
    return data


def test1(webapp, gziptool):
    request = Request(webapp.server.http.base)
    request.add_header("Accept-Encoding", "gzip")
    opener = build_opener()

    f = opener.open(request)
    s = decompress(f.read())
    assert s == b"Hello World!"


def test2(webapp, gziptool):
    request = Request("%s/static/largefile.txt" % webapp.server.http.base)
    request.add_header("Accept-Encoding", "gzip")
    opener = build_opener()

    f = opener.open(request)
    s = decompress(f.read())
    assert s == open(path.join(DOCROOT, "largefile.txt"), "rb").read()



# Module:   serial
# Date:     4th August 2004
# Author:   James Mills <prologic@shortcircuit.net.au>

"""Serial I/O

This module implements basic Serial (RS232) I/O.
"""

import os
import select
from collections import deque

from circuits.core import Component, handler, Event
from circuits.core.pollers import BasePoller, Poller
from circuits.core.utils import findcmp
from circuits.tools import tryimport
from circuits.six import binary_type, string_types

from .events import closed, error, opened, read, ready, close

serial = tryimport("serial")

TIMEOUT = 0.2
BUFSIZE = 4096


class _open(Event):
    """_open Event"""


class Serial(Component):

    channel = "serial"

    def __init__(self, port, baudrate=115200, bufsize=BUFSIZE,
                 timeout=TIMEOUT, channel=channel):
        super(Serial, self).__init__(channel=channel)

        if serial is None:
            raise RuntimeError("No serial support available")

        self._port = port
        self._baudrate = baudrate
        self._bufsize = bufsize

        self._serial = None
        self._poller = None
        self._buffer = deque()
        self._closeflag = False

    @handler("ready")
    def _on_ready(self, component):
        self.fire(_open(), self.channel)

    @handler("_open")
    def _on_open(self, port=None, baudrate=None, bufsize=None):
        self._port = port or self._port
        self._baudrate = baudrate or self._baudrate
        self._bufsize = bufsize or self._bufsize

        self._serial = serial.Serial(port=self._port, baudrate=self._baudrate, timeout=0)
        self._fd = self._serial.fileno()  # not portable!

        self._poller.addReader(self, self._fd)

        self.fire(opened(self._port, self._baudrate))

    @handler("registered", "started", channel="*")
    def _on_registered_or_started(self, component, manager=None):
        if self._poller is None:
            if isinstance(component, BasePoller):
                self._poller = component
                self.fire(ready(self))
            else:
                if component is not self:
                    return
                component = findcmp(self.root, BasePoller)
                if component is not None:
                    self._poller = component
                    self.fire(ready(self))
                else:
                    self._poller = Poller().register(self)
                    self.fire(ready(self))

    @handler("stopped", channel="*")
    def _on_stopped(self, component):
        self.fire(close())

    @handler("prepare_unregister", channel="*")
    def _on_prepare_unregister(self, event, c):
        if event.in_subtree(self):
            self._close()

    def _close(self):
        if self._closeflag:
            return

        self._poller.discard(self._fd)

        self._buffer.clear()
        self._closeflag = False
        self._connected = False

        try:
            self._serial.close()
        except:
            pass

        self.fire(closed())

    def close(self):
        if not self._buffer:
            self._close()
        elif not self._closeflag:
            self._closeflag = True

    def _read(self):
        try:
            data = self._serial.read(self._bufsize)
            if not isinstance(data, binary_type):
                data = data.encode(self._encoding)

            if data:
                self.fire(read(data)).notify = True
        except (OSError, IOError) as e:
            self.fire(error(e))
            self._close()

    def _write(self, data):
        try:
            if not isinstance(data, binary_type):
                data = data.encode(self._encoding)

            try:
                nbytes = self._serial.write(data)
            except (serial.SerialTimeoutException) as e:
                nbytes = 0
            if nbytes < len(data):
                self._buffer.appendleft(data[nbytes:])
        except (OSError, IOError) as e:
            self.fire(error(e))
            self._close()

    def write(self, data):
        if self._poller is not None and not self._poller.isWriting(self._fd):
            self._poller.addWriter(self, self._fd)
        self._buffer.append(data)

    @handler("_disconnect")
    def __on_disconnect(self, sock):
        self._close()

    @handler("_read")
    def __on_read(self, sock):
        self._read()

    @handler("_write")
    def __on_write(self, sock):
        if self._buffer:
            data = self._buffer.popleft()
            self._write(data)

        if not self._buffer:
            if self._closeflag:
                self._close()
            elif self._poller.isWriting(self._fd):
                self._poller.removeWriter(self._fd)



#!/usr/bin/env python

import pytest


from circuits.six import b, u
from circuits.web import Controller


from .helpers import urlencode, urlopen, HTTPError


class Root(Controller):

    def index(self):
        return "Hello World!"

    def test_args(self, *args, **kwargs):
        return "{0}\n{1}".format(repr(args), repr(kwargs))

    def test_default_args(self, a=None, b=None):
        return "a={0}\nb={1}".format(a, b)

    def test_redirect(self):
        return self.redirect("/")

    def test_forbidden(self):
        return self.forbidden()

    def test_notfound(self):
        return self.notfound()

    def test_failure(self):
        raise Exception()


def test_root(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"


def test_404(webapp):
    try:
        urlopen("%s/foo" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == "Not Found"
    else:
        assert False


def test_args(webapp):
    args = ("1", "2", "3")
    kwargs = {"1": "one", "2": "two", "3": "three"}
    url = "%s/test_args/%s" % (webapp.server.http.base, "/".join(args))
    data = urlencode(kwargs).encode('utf-8')
    f = urlopen(url, data)
    data = f.read().split(b"\n")
    assert eval(data[0]) == args
    assert eval(data[1]) == kwargs


@pytest.mark.parametrize("data,expected", [
    ((["1"], {}),           b("a=1\nb=None")),
    ((["1", "2"], {}),      b("a=1\nb=2")),
    ((["1"], {"b": "2"}),   b("a=1\nb=2")),
])
def test_default_args(webapp, data, expected):
    args, kwargs = data
    url = u("{0:s}/test_default_args/{1:s}".format(
        webapp.server.http.base,
        u("/").join(args)
    ))
    data = urlencode(kwargs).encode("utf-8")
    f = urlopen(url, data)
    assert f.read() == expected


def test_redirect(webapp):
    f = urlopen("%s/test_redirect" % webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"


def test_forbidden(webapp):
    try:
        urlopen("%s/test_forbidden" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 403
        assert e.msg == "Forbidden"
    else:
        assert False


def test_notfound(webapp):
    try:
        urlopen("%s/test_notfound" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == "Not Found"
    else:
        assert False


def test_failure(webapp):
    try:
        urlopen("%s/test_failure" % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 500
    else:
        assert False



#!/usr/bin/env python


"""Simple TCP Echo Server

This example shows how you can create a simple TCP Server (an Echo Service)
utilizing the builtin Socket Components that the circuits library ships with.
"""


from circuits import handler, Debugger
from circuits.net.sockets import TCPServer


class EchoServer(TCPServer):

    @handler("read")
    def on_read(self, sock, data):
        """Read Event Handler

        This is fired by the underlying Socket Component when there has been
        new data read from the connected client.

        ..note :: By simply returning, client/server socket components listen
                  to ValueChagned events (feedback) to determine if a handler
                  returned some data and fires a subsequent Write event with
                  the value returned.
        """

        return data


# Start and "run" the system.
# Bind to port 0.0.0.0:8000
app = EchoServer(("0.0.0.0", 8000))
Debugger().register(app)
app.run()



#!/usr/bin/env python

import pytest

from circuits import Component, Event


class App(Component):

    def init(self):
        self._ready = False
        self._done = False
        self._counter = 0

    def registered(self, component, manager):
        if component is self:
            self.fire(Event.create("ready"))

    def generate_events(self, event):
        if not self._ready or self._done:
            return

        if self._counter < 10:
            self.fire(Event.create("hello"))
        else:
            self.fire(Event.create("done"))
        event.reduce_time_left(0)

    def done(self):
        self._done = True

    def hello(self):
        self._counter += 1

    def ready(self):
        self._ready = True


@pytest.fixture(scope="module")
def app(request, manager, watcher):
    app = App().register(manager)

    def finalizer():
        app.unregister()

    request.addfinalizer(finalizer)

    assert watcher.wait("ready")

    return app


def test(manager, watcher, app):
    watcher.wait("done")
    assert app._counter == 10



#!/usr/bin/python -i

from circuits import Event, Component, Manager


class test(Event):
    """test Event"""


class App(Component):

    def test(self):
        return "Hello World!"

    def unregistered(self, *args):
        return

    def prepare_unregister(self, *args):
        return


m = Manager()
app = App()
app.register(m)

while app:
    app.flush()


def test_fire():
    x = m.fire(test())
    m.flush()
    assert x.value == "Hello World!"


def test_contains():
    assert App in m
    assert not m in app



#!/usr/bin/env python

from json import dumps

from circuits.web import Server, Controller


def json(f):
    def wrapper(self, *args, **kwargs):
        return dumps(f(self, *args, **kwargs))
    return wrapper


class Root(Controller):

    @json
    def getrange(self, limit=4):
        return list(range(int(limit)))

app = Server(("0.0.0.0", 8000))
Root().register(app)
app.run()



# Package:  net
# Date:     7th April 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Networking Components

This package contains components that implement network sockets and
protocols for implementing client and server network applications.

:copyright: CopyRight (C) 2004-2013 by James Mills
:license: MIT (See: LICENSE)
"""



# Package:  utils
# Date:     ...
# Author:   ...

"""Utils

...
"""


import json

from circuits.core import Event
from circuits.six import bytes_to_str, text_type


def load_event(s):
    data = json.loads(s)

    name = bytes_to_str(data["name"].encode("utf-8"))

    args = []
    for arg in data["args"]:
        if isinstance(arg, text_type):
            arg = arg.encode("utf-8")
        args.append(arg)

    kwargs = {}
    for k, v in data["kwargs"].items():
        if isinstance(v, text_type):
            v = v.encode("utf-8")
        kwargs[str(k)] = v

    e = Event.create(name, *args, **kwargs)

    e.success = bool(data["success"])
    e.failure = bool(data["failure"])
    e.notify = bool(data["notify"])
    e.channels = tuple(data["channels"])

    return e, data["id"]


def dump_event(e, id):
    data = {
        "id": id,
        "name": e.name,
        "args": e.args,
        "kwargs": e.kwargs,
        "success": e.success,
        "failure": e.failure,
        "channels": e.channels,
        "notify": e.notify
    }

    return json.dumps(data)


def dump_value(v):
    data = {
        "id": v.node_trn,
        "errors": v.errors,
        "value": v._value,
    }
    return json.dumps(data)


def load_value(v):
    data = json.loads(v)
    return data['value'], data['id'], data['errors']



#!/usr/bin/env python


"""Example IRC Server

.. note:: This is an example only and is feature incomplete.

Implements commands::

    USER NICK JOIN PART NICK WHO QUIT
"""


import logging
from logging import getLogger

from time import time
from sys import stderr
from itertools import chain
from operator import attrgetter
from collections import defaultdict
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser


from circuits import handler, Component, Debugger

from circuits.net.sockets import TCPServer
from circuits.net.events import close, write

from circuits.protocols.irc import joinprefix, reply, response, IRC, Message

from circuits.protocols.irc.replies import (
    ERR_NOMOTD, ERR_NOSUCHNICK, ERR_NOSUCHCHANNEL, ERR_UNKNOWNCOMMAND,
    RPL_WELCOME, RPL_YOURHOST, RPL_WHOREPLY, RPL_ENDOFWHO, RPL_NOTOPIC,
    RPL_NAMEREPLY, RPL_ENDOFNAMES,
    ERR_NICKNAMEINUSE,
)


__version__ = "0.0.1"


def parse_args():
    parser = ArgumentParser(
        description=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "-v", "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__)
    )

    parser.add_argument(
        "-b", "--bind",
        action="store", type=str,
        default="0.0.0.0:6667", dest="bind",
        help="Bind to address:[port]"
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        default=False, dest="debug",
        help="Enable debug mode"
    )

    return parser.parse_args()


class Channel(object):

    def __init__(self, name):
        self.name = name

        self.users = []


class User(object):

    def __init__(self, sock, host, port):
        self.sock = sock
        self.host = host
        self.port = port

        self.nick = None
        self.away = False
        self.channels = []
        self.signon = None
        self.registered = False
        self.userinfo = UserInfo()

    @property
    def prefix(self):
        userinfo = self.userinfo
        return joinprefix(self.nick, userinfo.user, userinfo.host)


class UserInfo(object):

    def __init__(self, user=None, host=None, name=None):
        self.user = user
        self.host = host
        self.name = name


class Server(Component):

    channel = "server"

    network = "Test"
    host = "localhost"
    version = "ircd v{0:s}".format(__version__)

    def init(self, args, logger=None):
        self.args = args
        self.logger = logger or getLogger(__name__)

        self.buffers = defaultdict(bytes)

        self.nicks = {}
        self.users = {}
        self.channels = {}

        Debugger(events=args.debug, logger=self.logger).register(self)

        if ":" in args.bind:
            address, port = args.bind.split(":")
            port = int(port)
        else:
            address, port = args.bind, 6667

        bind = (address, port)

        self.transport = TCPServer(
            bind,
            channel=self.channel
        ).register(self)

        self.protocol = IRC(
            channel=self.channel,
            getBuffer=self.buffers.__getitem__,
            updateBuffer=self.buffers.__setitem__
        ).register(self)

    def _notify(self, users, message, exclude=None):
        for user in users:
            if exclude is not None and user is exclude:
                continue
            self.fire(reply(user.sock, message))

    def read(self, sock, data):
        user = self.users[sock]
        host, port = user.host, user.port

        self.logger.info(
            "I: [{0:s}:{1:d}] {2:s}".format(host, port, repr(data))
        )

    def write(self, sock, data):
        user = self.users[sock]
        host, port = user.host, user.port

        self.logger.info(
            "O: [{0:s}:{1:d}] {2:s}".format(host, port, repr(data))
        )

    def ready(self, server, bind):
        stderr.write(
            "ircd v{0:s} ready! Listening on: {1:s}\n".format(
                __version__, "{0:s}:{1:d}".format(*bind)
            )
        )

    def connect(self, sock, host, port):
        self.users[sock] = User(sock, host, port)

        self.logger.info("C: [{0:s}:{1:d}]".format(host, port))

    def disconnect(self, sock):
        if sock not in self.users:
            return

        user = self.users[sock]

        self.logger.info("D: [{0:s}:{1:d}]".format(user.host, user.port))

        nick = user.nick
        user, host = user.userinfo.user, user.userinfo.host

        yield self.call(
            response.create("quit", sock, (nick, user, host), "Leavling")
        )

        del self.users[sock]

        if nick in self.nicks:
            del self.nicks[nick]

    def quit(self, sock, source, reason="Leaving"):
        user = self.users[sock]

        channels = [self.channels[channel] for channel in user.channels]
        for channel in channels:
            channel.users.remove(user)
            if not channel.users:
                del self.channels[channel.name]

        users = chain(*map(attrgetter("users"), channels))

        self.fire(close(sock))

        self._notify(
            users,
            Message("QUIT", reason, prefix=user.prefix), user
        )

    def nick(self, sock, source, nick):
        user = self.users[sock]

        if nick in self.nicks:
            return self.fire(reply(sock, ERR_NICKNAMEINUSE(nick)))

        if not user.registered:
            user.registered = True
            self.fire(response.create("signon", sock, user))

        user.nick = nick
        self.nicks[nick] = user

    def user(self, sock, source, nick, user, host, name):
        _user = self.users[sock]

        _user.userinfo.user = user
        _user.userinfo.host = host
        _user.userinfo.name = name

        if _user.nick is not None:
            _user.registered = True
            self.fire(response.create("signon", sock, source))

    def signon(self, sock, source):
        user = self.users[sock]
        if user.signon:
            return

        user.signon = time()

        self.fire(reply(sock, RPL_WELCOME(self.network)))
        self.fire(reply(sock, RPL_YOURHOST(self.host, self.version)))
        self.fire(reply(sock, ERR_NOMOTD()))

        # Force users to join #circuits
        self.fire(response.create("join", sock, source, "#circuits"))

    def join(self, sock, source, name):
        user = self.users[sock]

        if name not in self.channels:
            channel = self.channels[name] = Channel(name)
        else:
            channel = self.channels[name]

        if user in channel.users:
            return

        user.channels.append(name)
        channel.users.append(user)

        self._notify(
            channel.users,
            Message("JOIN", name, prefix=user.prefix)
        )

        self.fire(reply(sock, RPL_NOTOPIC(name)))
        self.fire(reply(sock, RPL_NAMEREPLY(channel)))
        self.fire(reply(sock, RPL_ENDOFNAMES()))

    def part(self, sock, source, name, reason="Leaving"):
        user = self.users[sock]

        channel = self.channels[name]

        self._notify(
            channel.users,
            Message("PART", name, reason, prefix=user.prefix)
        )

        user.channels.remove(name)
        channel.users.remove(user)

        if not channel.users:
            del self.channels[name]

    def privmsg(self, sock, source, target, message):
        user = self.users[sock]

        if target.startswith("#"):
            if target not in self.channels:
                return self.fire(reply(sock, ERR_NOSUCHCHANNEL(target)))

            channel = self.channels[target]

            self._notify(
                channel.users,
                Message("PRIVMSG", target, message, prefix=user.prefix),
                user
            )
        else:
            if target not in self.nicks:
                return self.fire(reply(sock, ERR_NOSUCHNICK(target)))

            self.fire(
                reply(
                    self.nicks[target].sock,
                    Message("PRIVMSG", target, message, prefix=user.prefix)
                )
            )

    def who(self, sock, source, mask):
        if mask.startswith("#"):
            if mask not in self.channels:
                return self.fire(reply(sock, ERR_NOSUCHCHANNEL(mask)))

            channel = self.channels[mask]

            for user in channel.users:
                self.fire(reply(sock, RPL_WHOREPLY(user, mask, self.host)))
            self.fire(reply(sock, RPL_ENDOFWHO(mask)))
        else:
            if mask not in self.nicks:
                return self.fire(reply(sock, ERR_NOSUCHNICK(mask)))

            user = self.nicks[mask]

            self.fire(reply(sock, RPL_WHOREPLY(user, mask, self.host)))
            self.fire(reply(sock, RPL_ENDOFWHO(mask)))

    def ping(self, event, sock, source, server):
        event.stop()
        self.fire(reply(sock, Message("PONG", server)))

    def reply(self, target, message):
        user = self.users[target]

        if message.add_nick:
            message.args.insert(0, user.nick or "")

        if message.prefix is None:
            message.prefix = self.host

        self.fire(write(target, bytes(message)))

    @property
    def commands(self):
        exclude = {"ready", "connect", "disconnect", "read", "write"}
        return list(set(self.events()) - exclude)

    @handler()
    def _on_event(self, event, *args, **kwargs):
        if event.name.endswith("_done"):
            return

        if isinstance(event, response):
            if event.name not in self.commands:
                event.stop()
                self.fire(reply(args[0], ERR_UNKNOWNCOMMAND(event.name)))


def main():
    args = parse_args()

    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        stream=stderr,
        level=logging.DEBUG if args.debug else logging.INFO
    )

    logger = getLogger(__name__)

    Server(args, logger=logger).run()


if __name__ == "__main__":
    main()



# Module:   errors
# Date:     11th February 2009
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Errors

This module implements a set of standard HTTP Errors.
"""

from cgi import escape

try:
    from urllib.parse import urljoin as _urljoin
except ImportError:
    from urlparse import urljoin as _urljoin  # NOQA

from circuits import Event

from ..six import string_types
from .constants import SERVER_URL, SERVER_VERSION
from .constants import DEFAULT_ERROR_MESSAGE, HTTP_STATUS_CODES


class httperror(Event):
    """An event for signaling an HTTP error"""

    code = 500
    description = ""
    contenttype = "text/html"

    def __init__(self, request, response, code=None, **kwargs):
        """
        The constructor creates a new instance and modifies the *response*
        argument to reflect the error.
        """
        super(httperror, self).__init__(request, response, code, **kwargs)

        # Override HTTPError subclasses
        self.name = "httperror"

        self.request = request
        self.response = response

        if code is not None:
            self.code = code

        self.error = kwargs.get("error", None)

        self.description = kwargs.get(
            "description", getattr(self.__class__, "description", "")
        )

        if self.error is not None:
            self.traceback = "ERROR: (%s) %s\n%s" % (
                self.error[0], self.error[1], "".join(self.error[2])
            )
        else:
            self.traceback = ""

        self.response.close = True
        self.response.status = self.code

        self.response.headers["Content-Type"] = self.contenttype

        self.data = {
            "code": self.code,
            "name": HTTP_STATUS_CODES.get(self.code, "???"),
            "description": self.description,
            "traceback": escape(self.traceback),
            "url": SERVER_URL,
            "version": SERVER_VERSION
        }

    def sanitize(self):
        if self.code != 201 and not (299 < self.code < 400):
            if "Location" in self.response.headers:
                del self.response.headers["Location"]

    def __str__(self):
        self.sanitize()
        return DEFAULT_ERROR_MESSAGE % self.data

    def __repr__(self):
        return "<%s %d %s>" % (
            self.__class__.__name__, self.code, HTTP_STATUS_CODES.get(
                self.code, "???"
            )
        )


class forbidden(httperror):
    """An event for signaling the HTTP Forbidden error"""

    code = 403


class unauthorized(httperror):
    """An event for signaling the HTTP Unauthorized error"""

    code = 401


class notfound(httperror):
    """An event for signaling the HTTP Not Fouond error"""

    code = 404


class redirect(httperror):
    """An event for signaling the HTTP Redirect response"""

    def __init__(self, request, response, urls, code=None):
        """
        The constructor creates a new instance and modifies the
        *response* argument to reflect a redirect response to the
        given *url*.
        """

        if isinstance(urls, string_types):
            urls = [urls]

        abs_urls = []
        for url in urls:
            # Note that urljoin will "do the right thing" whether url is:
            #  1. a complete URL with host (e.g. "http://www.example.com/test")
            #  2. a URL relative to root (e.g. "/dummy")
            #  3. a URL relative to the current path
            # Note that any query string in request is discarded.
            url = request.uri.relative(url).unicode()
            abs_urls.append(url)
        self.urls = urls = abs_urls

        # RFC 2616 indicates a 301 response code fits our goal; however,
        # browser support for 301 is quite messy. Do 302/303 instead. See
        # http://ppewww.ph.gla.ac.uk/~flavell/www/post-redirect.html
        if code is None:
            if request.protocol >= (1, 1):
                code = 303
            else:
                code = 302
        else:
            if code < 300 or code > 399:
                raise ValueError("status code must be between 300 and 399.")

        super(redirect, self).__init__(request, response, code)

        if code in (300, 301, 302, 303, 307):
            response.headers["Content-Type"] = "text/html"
            # "The ... URI SHOULD be given by the Location field
            # in the response."
            response.headers["Location"] = urls[0]

            # "Unless the request method was HEAD, the entity of the response
            # SHOULD contain a short hypertext note with a hyperlink to the
            # new URI(s)."
            msg = {300: "This resource can be found at <a href='%s'>%s</a>.",
                   301: ("This resource has permanently moved to "
                         "<a href='%s'>%s</a>."),
                   302: ("This resource resides temporarily at "
                         "<a href='%s'>%s</a>."),
                   303: ("This resource can be found at "
                         "<a href='%s'>%s</a>."),
                   307: ("This resource has moved temporarily to "
                         "<a href='%s'>%s</a>."),
                   }[code]
            response.body = "<br />\n".join([msg % (u, u) for u in urls])
            # Previous code may have set C-L, so we have to reset it
            # (allow finalize to set it).
            response.headers.pop("Content-Length", None)
        elif code == 304:
            # Not Modified.
            # "The response MUST include the following header fields:
            # Date, unless its omission is required by section 14.18.1"
            # The "Date" header should have been set in Response.__init__

            # "...the response SHOULD NOT include other entity-headers."
            for key in ("Allow", "Content-Encoding", "Content-Language",
                        "Content-Length", "Content-Location", "Content-MD5",
                        "Content-Range", "Content-Type", "Expires",
                        "Last-Modified"):
                if key in response.headers:
                    del response.headers[key]

            # "The 304 response MUST NOT contain a message-body."
            response.body = None
            # Previous code may have set C-L, so we have to reset it.
            response.headers.pop("Content-Length", None)
        elif code == 305:
            # Use Proxy.
            # urls[0] should be the URI of the proxy.
            response.headers["Location"] = urls[0]
            response.body = None
            # Previous code may have set C-L, so we have to reset it.
            response.headers.pop("Content-Length", None)
        else:
            raise ValueError("The %s status code is unknown." % code)

    def __repr__(self):
        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ""
        return "<%s %d[%s.%s] %s>" % (
            self.__class__.__name__, self.code, channels, self.name,
            " ".join(self.urls)
        )



#!/usr/bin/env python

import pytest

from circuits import handler, Event, Component


class test(Event):
    """test Event"""


class Base(Component):

    def test(self):
        return "Hello World!"


class App1(Base):

    @handler("test", priority=-1)
    def test(self):
        return "Foobar"


class App2(Base):

    @handler("test", override=True)
    def test(self):
        return "Foobar"


def test_inheritence():
    app = App1()
    app.start()

    x = app.fire(test())
    assert pytest.wait_for(x, "result")
    v = x.value
    assert v == ["Hello World!", "Foobar"]

    app.stop()


def test_override():
    app = App2()
    app.start()

    x = app.fire(test())
    assert pytest.wait_for(x, "result")
    v = x.value
    assert v == "Foobar"

    app.stop()



# Package:  parsers
# Date:     26th March 2013
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""circuits.web parsers"""

from .multipart import MultipartParser
from .querystring import QueryStringParser
from .http import HttpParser, BAD_FIRST_LINE

# flake8: noqa



# Module:   file
# Date:     4th August 2004
# Author:   James Mills <prologic@shortcircuit.net.au>

"""File I/O

This module implements a wrapper for basic File I/O.
"""

try:
    from os import O_NONBLOCK
except ImportError:
    # If it fails, that's fine. the fcntl import
    # will fail anyway.
    pass

from collections import deque
from os import read as fd_read
from os import write as fd_write
from sys import getdefaultencoding
from errno import EINTR, EWOULDBLOCK

from circuits.tools import tryimport
from circuits.core.utils import findcmp
from circuits.core import handler, Component, Event
from circuits.core.pollers import BasePoller, Poller
from circuits.six import binary_type, string_types, PY3

from .events import close, closed, eof, error, opened, read, ready

fcntl = tryimport("fcntl")

TIMEOUT = 0.2
BUFSIZE = 4096


class _open(Event):
    """_open Event"""


class File(Component):

    channel = "file"

    def init(self, filename, mode="r", bufsize=BUFSIZE, encoding=None,
             channel=channel):
        self._mode = mode
        self._bufsize = bufsize
        self._filename = filename
        self._encoding = encoding or getdefaultencoding()

        self._fd = None
        self._poller = None
        self._buffer = deque()
        self._closeflag = False

    @property
    def closed(self):
        return getattr(self._fd, "closed", True) \
            if hasattr(self, "_fd") else True

    @property
    def filename(self):
        return getattr(self, "_filename", None)

    @property
    def mode(self):
        return getattr(self, "_mode", None)

    @handler("ready")
    def _on_ready(self, component):
        self.fire(_open(), self.channel)

    @handler("_open")
    def _on_open(self, filename=None, mode=None, bufsize=None):
        self._filename = filename or self._filename
        self._bufsize = bufsize or self._bufsize
        self._mode = mode or self._mode

        if isinstance(self._filename, string_types[0]):
            kwargs = {"encoding": self._encoding} if PY3 else {}
            self._fd = open(self.filename, self.mode, **kwargs)
        else:
            self._fd = self._filename
            self._mode = self._fd.mode
            self._filename = self._fd.name
            self._encoding = getattr(self._fd, "encoding", self._encoding)

        if fcntl is not None:
            # Set non-blocking file descriptor (non-portable)
            flag = fcntl.fcntl(self._fd, fcntl.F_GETFL)
            flag = flag | O_NONBLOCK
            fcntl.fcntl(self._fd, fcntl.F_SETFL, flag)

        if "r" in self.mode or "+" in self.mode:
            self._poller.addReader(self, self._fd)

        self.fire(opened(self.filename, self.mode))

    @handler("registered", "started", channel="*")
    def _on_registered_or_started(self, component, manager=None):
        if self._poller is None:
            if isinstance(component, BasePoller):
                self._poller = component
                self.fire(ready(self))
            else:
                if component is not self:
                    return
                component = findcmp(self.root, BasePoller)
                if component is not None:
                    self._poller = component
                    self.fire(ready(self))
                else:
                    self._poller = Poller().register(self)
                    self.fire(ready(self))

    @handler("stopped", channel="*")
    def _on_stopped(self, component):
        self.fire(close())

    @handler("prepare_unregister", channel="*")
    def _on_prepare_unregister(self, event, c):
        if event.in_subtree(self):
            self._close()

    def _close(self):
        if self.closed:
            return

        self._poller.discard(self._fd)

        self._buffer.clear()
        self._closeflag = False
        self._connected = False

        try:
            self._fd.close()
        except:
            pass

        self.fire(closed())

    def close(self):
        if not self._buffer:
            self._close()
        elif not self._closeflag:
            self._closeflag = True

    def _read(self):
        try:
            data = fd_read(self._fd.fileno(), self._bufsize)
            if not isinstance(data, binary_type):
                data = data.encode(self._encoding)

            if data:
                self.fire(read(data)).notify = True
            else:
                self.fire(eof())
                if not any(m in self.mode for m in ("a", "+")):
                    self.close()
                else:
                    self._poller.discard(self._fd)
        except (OSError, IOError) as e:
            if e.args[0] in (EWOULDBLOCK, EINTR):
                return
            else:
                self.fire(error(e))
                self._close()

    def seek(self, offset, whence=0):
        self._fd.seek(offset, whence)

    def _write(self, data):
        try:
            if not isinstance(data, binary_type):
                data = data.encode(self._encoding)

            nbytes = fd_write(self._fd.fileno(), data)

            if nbytes < len(data):
                self._buffer.appendleft(data[nbytes:])
        except (OSError, IOError) as e:
            if e.args[0] in (EWOULDBLOCK, EINTR):
                return
            else:
                self.fire(error(e))
                self._close()

    def write(self, data):
        if self._poller is not None and not self._poller.isWriting(self._fd):
            self._poller.addWriter(self, self._fd)
        self._buffer.append(data)

    @handler("_disconnect")
    def __on_disconnect(self, sock):
        self._close()

    @handler("_read")
    def __on_read(self, sock):
        self._read()

    @handler("_write")
    def __on_write(self, sock):
        if self._buffer:
            data = self._buffer.popleft()
            self._write(data)

        if not self._buffer:
            if self._closeflag:
                self._close()
            elif self._poller.isWriting(self._fd):
                self._poller.removeWriter(self._fd)



#!/usr/bin/env python

import pytest

from os import path
from io import BytesIO

from circuits.web import Controller

from .multipartform import MultiPartForm
from .helpers import urlopen, Request


@pytest.fixture()
def sample_file(request):
    return open(
        path.join(
            path.dirname(__file__),
            "static", "unicode.txt"
        ),
        "rb"
    )


class Root(Controller):

    def index(self, file, description=""):
        yield "Filename: %s\n" % file.filename
        yield "Description: %s\n" % description
        yield "Content:\n"
        yield file.value

    def upload(self, file, description=""):
        return file.value


def test(webapp):
    form = MultiPartForm()
    form["description"] = "Hello World!"

    fd = BytesIO(b"Hello World!")
    form.add_file("file", "helloworld.txt", fd, "text/plain; charset=utf-8")

    # Build the request
    url = webapp.server.http.base
    data = form.bytes()
    headers = {
        "Content-Type": form.get_content_type(),
        "Content-Length": len(data),
    }

    request = Request(url, data, headers)

    f = urlopen(request)
    s = f.read()
    lines = s.split(b"\n")

    assert lines[0] == b"Filename: helloworld.txt"
    assert lines[1] == b"Description: Hello World!"
    assert lines[2] == b"Content:"
    assert lines[3] == b"Hello World!"


def test_unicode(webapp, sample_file):
    form = MultiPartForm()
    form["description"] = sample_file.name
    form.add_file(
        "file", "helloworld.txt", sample_file,
        "text/plain; charset=utf-8"
    )

    # Build the request
    url = "{0:s}/upload".format(webapp.server.http.base)
    data = form.bytes()
    headers = {
        "Content-Type": form.get_content_type(),
        "Content-Length": len(data),
    }

    request = Request(url, data, headers)

    f = urlopen(request)
    s = f.read()
    sample_file.seek(0)
    expected_output = sample_file.read()  # use the byte stream
    assert s == expected_output



#!/usr/bin/env python

from datetime import datetime
from time import mktime
from email.utils import parsedate

from circuits.web import Controller

from .helpers import urlopen


class Root(Controller):

    def index(self):
        self.expires(60)
        return "Hello World!"

    def nocache(self):
        self.expires(0)
        return "Hello World!"


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"
    expires = f.headers["Expires"]
    diff = (mktime(parsedate(expires)) - mktime(datetime.utcnow().timetuple()))
    assert 60 - (60 * 0.1) < diff < 60 + (60 * 0.1)  # diff is about 60 +- 10%


def test_nocache(webapp):
    f = urlopen("%s/nocache" % webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    expires = f.headers["Expires"]
    pragma = f.headers["Pragma"]
    cacheControl = f.headers["Cache-Control"]

    now = datetime.utcnow()
    lastyear = now.replace(year=now.year-1)

    diff = (mktime(parsedate(expires)) - mktime(lastyear.utctimetuple()))
    assert diff < 1.0

    assert pragma == "no-cache"
    assert cacheControl == "no-cache, must-revalidate"



#!/usr/bin/env python

from circuits.web import Controller
from circuits.web.wsgi import Application


class Root(Controller):

    def index(self):
        return "Hello World!"

application = Application()
Root().register(application)



#!/usr/bin/env python

from circuits import handler, Event, Component


class foo(Event):
    """foo Event"""


class test(Event):
    """test Event"""


class A(Component):

    channel = "a"

    def test(self):
        return "Hello World!"

    @handler(priority=1.0)
    def _on_event(self, event, *args, **kwargs):
        return "Foo"


class B(Component):

    @handler(priority=10.0, channel="*")
    def _on_channel(self, event, *args, **kwargs):
        return "Bar"


def test_main():
    app = A() + B()
    while app:
        app.flush()

    x = app.fire(test(), "a")
    while app:
        app.flush()

    assert x.value[0] == "Bar"
    assert x.value[1] == "Foo"
    assert x.value[2] == "Hello World!"


def test_event():
    app = A() + B()
    while app:
        app.flush()

    e = test()
    x = app.fire(e)
    while app:
        app.flush()

    assert x.value[0] == "Bar"
    assert x.value[1] == "Foo"
    assert x.value[2] == "Hello World!"


def test_channel():
    app = A() + B()
    while app:
        app.flush()

    e = foo()
    x = app.fire(e, "b")
    while app:
        app.flush()

    assert x.value == "Bar"



#!/usr/bin/env python


import pytest


import os
import sys
from time import sleep
from errno import ESRCH
from signal import SIGTERM
from os import kill, remove
from subprocess import Popen


from . import signalapp


def is_running(pid):
    try:
        kill(pid, 0)
    except OSError as error:
        if error.errno == ESRCH:
            return False
    return True


def wait(pid, timeout=3):
    count = timeout
    while is_running(pid) and count:
        sleep(1)


def test(tmpdir):
    if not os.name == "posix":
        pytest.skip("Cannot run test on a non-POSIX platform.")

    tmpdir.ensure(".pid")
    tmpdir.ensure(".signal")
    pidfile = str(tmpdir.join(".pid"))
    signalfile = str(tmpdir.join(".signal"))

    args = [sys.executable, signalapp.__file__, pidfile, signalfile]
    cmd = " ".join(args)
    p = Popen(cmd, shell=True, env={'PYTHONPATH': ':'.join(sys.path)})
    status = p.wait()

    assert status == 0

    sleep(1)

    assert os.path.exists(pidfile)
    assert os.path.isfile(pidfile)

    f = open(pidfile, "r")
    pid = int(f.read().strip())
    f.close()

    kill(pid, SIGTERM)
    wait(pid)

    f = open(signalfile, "r")
    signal = f.read().strip()
    f.close()

    assert signal == str(SIGTERM)

    remove(pidfile)
    remove(signalfile)



#!/usr/bin/env python


from hashlib import md5


from circuits import handler, Component
from circuits.web import Server, Controller
from circuits.web.tools import check_auth, basic_auth


class Auth(Component):

    realm = "Test"
    users = {"admin": md5("admin").hexdigest()}

    @handler("request", priority=1.0)
    def on_request(self, event, request, response):
        """Filter Requests applying Basic Authentication

        Filter any incoming requests at a higher priority than the
        default dispatcher and apply Basic Authentication returning
        a 403 Forbidden response if Authentication failed.
        """

        if not check_auth(request, response, self.realm, self.users):
            event.stop()
            return basic_auth(request, response, self.realm, self.users)


class Root(Controller):

    def index(self):
        return "Hello World!"

app = Server(("0.0.0.0", 8000))
Auth().register(app)
Root().register(app)
app.run()




import itertools
from mimetypes import guess_type
from email.generator import _make_boundary


class MultiPartForm(dict):

    def __init__(self):
        self.files = []
        self.boundary = _make_boundary()

    def get_content_type(self):
        return "multipart/form-data; boundary=%s" % self.boundary

    def add_file(self, fieldname, filename, fd, mimetype=None):
        body = fd.read()
        if mimetype is None:
            mimetype = guess_type(filename)[0] or "application/octet-stream"
        self.files.append((fieldname, filename, mimetype, body))

    def bytes(self):
        parts = []
        part_boundary = bytearray("--%s" % self.boundary, "ascii")

        # Add the form fields
        parts.extend([
            part_boundary,
            bytearray(
                "Content-Disposition: form-data; name=\"%s\"" % k,
                "ascii"
            ),
            bytes(),
            v if isinstance(v, bytes) else bytearray(v, "ascii")
        ] for k, v in list(self.items()))

        # Add the files to upload
        parts.extend([
            part_boundary,
            bytearray(
                "Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"" % (
                    fieldname, filename),
                "ascii"
            ),
            bytearray("Content-Type: %s" % content_type, "ascii"),
            bytearray(),
            body if isinstance(body, bytes) else bytearray(body, "ascii"),
        ] for fieldname, filename, content_type, body in self.files)

        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data
        flattened = list(itertools.chain(*parts))
        flattened.append(bytearray("--%s--" % self.boundary, "ascii"))
        res = bytearray()
        for item in flattened:
            res += item
            res += bytearray("\r\n", "ascii")
        return res



#!/usr/bin/env python

from circuits import handler, BaseComponent, Debugger


class MyComponent(BaseComponent):

    def __init__(self):
        super(MyComponent, self).__init__()

        Debugger().register(self)

    @handler("started", channel="*")
    def system_started(self, component):
        print "Start event detected"

MyComponent().run()



#!/usr/bin/env python

from circuits.web import Controller
from circuits.web.client import Client, request


class Root(Controller):
    def __init__(self, *args, **kwargs):
        super(Root, self).__init__(*args, **kwargs)
        self += Leaf()

    def index(self):
        return "Hello World!"

    def name(self):
        return "Earth"


class Leaf(Controller):

    channel = "/world/country/region"

    def index(self):
        return "Hello cities!"

    def city(self):
        return "Hello City!"


def make_request(webapp, path):
    client = Client()
    client.start()

    client.fire(request("GET", path))
    while client.response is None:
        pass

    client.stop()

    response = client.response
    s = response.read()
    return response.status, s


def test_root(webapp):
    status, content = make_request(webapp, webapp.server.http.base)

    assert status == 200
    assert content == b"Hello World!"


def test_root_name(webapp):
    status, content = make_request(webapp, "%s/name" % webapp.server.http.base)

    assert status == 200
    assert content == b"Earth"


def test_leaf(webapp):
    status, content = make_request(webapp, "%s/world/country/region" % webapp.server.http.base)

    assert status == 200
    assert content == b"Hello cities!"


def test_city(webapp):
    status, content = make_request(webapp, "%s/world/country/region/city" % webapp.server.http.base)

    assert status == 200
    assert content == b"Hello City!"



# Module:   test_workers
# Date:     7th October 2008
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Workers Tests"""

import pytest

from circuits import task, Worker


@pytest.fixture(scope="module")
def worker(request):
    worker = Worker()

    def finalizer():
        worker.stop()

    request.addfinalizer(finalizer)

    if request.config.option.verbose:
        from circuits import Debugger
        Debugger().register(worker)

    waiter = pytest.WaitEvent(worker, "started")
    worker.start()
    assert waiter.wait()

    return worker


def f():
    x = 0
    i = 0
    while i < 1000000:
        x += 1
        i += 1
    return x


def add(a, b):
    return a + b


def test(worker):
    x = worker.fire(task(f))

    assert pytest.wait_for(x, "result")

    assert x.result
    assert x.value == 1000000


def test_args(worker):
    x = worker.fire(task(add, 1, 2))

    assert pytest.wait_for(x, "result")

    assert x.result
    assert x.value == 3



import re
from cgi import parse_header

from .headers import HeaderElement
from .parsers import MultipartParser
from .parsers import QueryStringParser


def process_multipart(request, params):
    headers = request.headers

    ctype = headers.elements("Content-Type")
    if ctype:
        ctype = ctype[0]
    else:
        ctype = HeaderElement.from_str("application/x-www-form-urlencoded")

    ib = ""
    if "boundary" in ctype.params:
        # http://tools.ietf.org/html/rfc2046#section-5.1.1
        # "The grammar for parameters on the Content-type field is such that it
        # is often necessary to enclose the boundary parameter values in quotes
        # on the Content-type line"
        ib = ctype.params["boundary"].strip("\"")

    if not re.match("^[ -~]{0,200}[!-~]$", ib):
        raise ValueError("Invalid boundary in multipart form: %r" % (ib,))

    parser = MultipartParser(request.body, ib)
    for part in parser:
        if part.filename or not part.is_buffered():
            params[part.name] = part
        else:
            params[part.name] = part.value


def process_urlencoded(request, params, encoding="utf-8"):
    params.update(QueryStringParser(request.qs).result)
    body = request.body.getvalue().decode(encoding)
    params.update(QueryStringParser(body).result)


def process(request, params):
    ctype = request.headers.get("Content-Type")
    if not ctype:
        return

    mtype, mencoding = ctype.split("/", 1) if "/" in ctype else (ctype, None)
    mencoding, extra = parse_header(mencoding)

    charset = extra.get("charset", "utf-8")

    if mtype == "multipart":
        process_multipart(request, params)
    elif mtype == "application" and mencoding == "x-www-form-urlencoded":
        process_urlencoded(request, params, encoding=charset)



#!/usr/bin/env python


from __future__ import print_function


from circuits import Component
from circuits.web.servers import Server
from circuits.web.controllers import Controller
from circuits.net.sockets import close, write
from circuits.web.websockets import WebSocketClient, WebSocketsDispatcher


from .helpers import urlopen


class Echo(Component):

    channel = "wsserver"

    def init(self):
        self.clients = []

    def connect(self, sock, host, port):
        self.clients.append(sock)
        print("WebSocket Client Connected:", host, port)
        self.fire(write(sock, "Welcome {0:s}:{1:d}".format(host, port)))

    def disconnect(self, sock):
        self.clients.remove(sock)

    def read(self, sock, data):
        self.fire(write(sock, "Received: " + data))


class Root(Controller):

    def index(self):
        return "Hello World!"


class Client(Component):

    channel = "ws"

    def init(self, *args, **kwargs):
        self.response = None

    def read(self, data):
        self.response = data


def test(manager, watcher, webapp):
    server = Server(0).register(manager)
    watcher.wait("ready")

    echo = Echo().register(server)
    Root().register(server)
    watcher.wait("registered", channel="wsserver")

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    watcher.clear()

    WebSocketsDispatcher("/websocket").register(server)
    watcher.wait("registered", channel="web")

    uri = "ws://{0:s}:{1:d}/websocket".format(server.host, server.port)

    WebSocketClient(uri).register(manager)
    client = Client().register(manager)
    watcher.wait("registered", channel="wsclient")
    watcher.wait("connected", channel="wsclient")

    assert len(echo.clients) == 1

    watcher.wait("read", channel="ws")
    assert client.response.startswith("Welcome")
    watcher.clear()

    client.fire(write("Hello!"), "ws")
    watcher.wait("read", channel="ws")
    assert client.response == "Received: Hello!"

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    assert len(echo.clients) == 1

    client.fire(close(), "ws")
    watcher.wait("disconnect", channel="wsserver")
    assert len(echo.clients) == 0

    client.unregister()
    watcher.wait("unregistered")
    watcher.clear()

    server.unregister()
    watcher.wait("unregistered")



# Module:   debugger
# Date:     5th November 2008
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Debugger Tests"""

import sys

import pytest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO  # NOQA

from circuits import Debugger
from circuits.core import Event, Component


class test(Event):
    """test Event"""


class App(Component):

    def test(self, raiseException=False):
        if raiseException:
            raise Exception()


class Logger(object):

    error_msg = None
    debug_msg = None

    def error(self, msg):
        self.error_msg = msg

    def debug(self, msg):
        self.debug_msg = msg


def test_main():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while app:
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    e = Event()
    app.fire(e)
    app.flush()
    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    assert not debugger._events

    e = Event()
    app.fire(e)

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""
    stderr.seek(0)
    stderr.truncate()


def test_file(tmpdir):
    logfile = str(tmpdir.ensure("debug.log"))
    stderr = open(logfile, "w+")

    app = App()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while app:
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    assert not debugger._events

    e = Event()
    app.fire(e)

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""
    stderr.seek(0)
    stderr.truncate()


def test_filename(tmpdir):
    if "__pypy__" in sys.modules:
        pytest.skip("Broken on pypy")

    logfile = str(tmpdir.ensure("debug.log"))
    stderr = open(logfile, "r+")

    app = App()
    debugger = Debugger(file=logfile)
    debugger.register(app)
    while app:
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    assert not debugger._events

    e = Event()
    app.fire(e)

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""
    stderr.seek(0)
    stderr.truncate()


def test_exceptions():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while app:
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events
    assert debugger._errors

    e = test(raiseException=True)
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    app.flush()
    stderr.seek(0)
    s = stderr.read().strip()
    assert s.startswith("<exception[*]")
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    debugger._errors = False

    assert not debugger._events
    assert not debugger._errors

    e = test(raiseException=True)
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""
    stderr.seek(0)
    stderr.truncate()

    app.flush()
    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""


def test_IgnoreEvents():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while app:
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    debugger.IgnoreEvents.extend(["test"])

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    e = test()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""
    stderr.seek(0)
    stderr.truncate()


def test_IgnoreChannels():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while app:
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events
    debugger.IgnoreChannels.extend([("*", "test")])

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    e = test()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ""
    stderr.seek(0)
    stderr.truncate()


def test_Logger_debug():
    app = App()
    logger = Logger()
    debugger = Debugger(logger=logger)
    debugger.register(app)
    while app:
        app.flush()

    e = Event()
    app.fire(e)
    app.flush()

    assert logger.debug_msg == repr(e)


def test_Logger_error():
    app = App()
    logger = Logger()
    debugger = Debugger(logger=logger)
    debugger.register(app)
    while app:
        app.flush()

    e = test(raiseException=True)
    app.fire(e)
    while app:
        app.flush()

    assert logger.error_msg.startswith("ERROR <handler[*.test] (App.test)> (")



#!/usr/bin/env python

from .helpers import urlopen

from circuits.web import Controller


class Root(Controller):

    def index(self, *args, **kwargs):
        return "ERROR"


def application(environ, start_response):
    status = "200 OK"
    start_response(status, [])
    return [""]


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b""



#!/usr/bin/env python


"""Chat Server Example

This example demonstrates how to create a very simple telnet-style chat
server that supports many connecting clients.
"""


from optparse import OptionParser


from circuits.net.events import write
from circuits import Component, Debugger
from circuits.net.sockets import TCPServer


__version__ = "0.0.1"

USAGE = "%prog [options]"
VERSION = "%prog v" + __version__


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        "-b", "--bind",
        action="store", type="string",
        default="0.0.0.0:8000", dest="bind",
        help="Bind to address:[port]"
    )

    parser.add_option(
        "-d", "--debug",
        action="store_true",
        default=False, dest="debug",
        help="Enable debug mode"
    )

    opts, args = parser.parse_args()

    return opts, args


class ChatServer(Component):

    def init(self, args, opts):
        """Initialize our ``ChatServer`` Component.

        This uses the convenience ``init`` method which is called after the
        component is proeprly constructed and initialized and passed the
        same args and kwargs that were passed during construction.
        """

        self.args = args
        self.opts = opts

        self.clients = {}

        if opts.debug:
            Debugger().register(self)

        if ":" in opts.bind:
            address, port = opts.bind.split(":")
            port = int(port)
        else:
            address, port = opts.bind, 8000

        bind = (address, port)

        TCPServer(bind).register(self)

    def broadcast(self, data, exclude=None):
        exclude = exclude or []
        targets = (sock for sock in self.clients.keys() if sock not in exclude)
        for target in targets:
            self.fire(write(target, data))

    def connect(self, sock, host, port):
        """Connect Event -- Triggered for new connecting clients"""

        self.clients[sock] = {
            "host": sock,
            "port": port,
            "state": {
                "nickname": None,
                "registered": False
            }
        }

        self.fire(write(sock, b"Welcome to the circuits Chat Server!\n"))
        self.fire(write(sock, b"Please enter a desired nickname: "))

    def disconnect(self, sock):
        """Disconnect Event -- Triggered for disconnecting clients"""

        if sock not in self.clients:
            return

        nickname = self.clients[sock]["state"]["nickname"]

        self.broadcast(
            "!!! {0:s} has left !!!\n".format(nickname).encode("utf-8"),
            exclude=[sock]
        )

        del self.clients[sock]

    def read(self, sock, data):
        """Read Event -- Triggered for when client conenctions have data"""

        data = data.strip().decode("utf-8")

        if not self.clients[sock]["state"]["registered"]:
            nickname = data
            self.clients[sock]["state"]["registered"] = True
            self.clients[sock]["state"]["nickname"] = nickname

            self.broadcast(
                "!!! {0:s} has joined !!!\n".format(nickname).encode("utf-8"),
                exclude=[sock]
            )
        else:
            nickname = self.clients[sock]["state"]["nickname"]

            self.broadcast(
                "<{0:s}> {1:s}\n".format(nickname, data).encode("utf-8"),
                exclude=[sock]
            )


def main():
    opts, args = parse_options()

    # Configure and "run" the System.
    ChatServer(args, opts).run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Test Interface Query

Test the capabilities of querying a Component class or instance for it's
interface. That is it's event handlers it responds to.
"""


from circuits import Component


class Base(Component):

    def foo(self):
        pass


class SuperBase(Base):

    def bar(self):
        pass


def test_handles_base_class():
    assert Base.handles("foo")


def test_handles_super_base_class():
    assert SuperBase.handles("foo", "bar")


def test_handles_base_instance():
    base = Base()
    assert base.handles("foo")


def test_handles_super_base_instance():
    superbase = SuperBase()
    assert superbase.handles("foo", "bar")



#!/usr/bin/env python

from collections import defaultdict

from circuits.protocols import Line
from circuits import Event, Component


class read(Event):
    """read Event"""


class App(Component):

    lines = []

    def line(self, line):
        self.lines.append(line)


class AppServer(Component):

    channel = "server"

    lines = []

    def line(self, sock, line):
        self.lines.append((sock, line))


def test():
    app = App()
    Line().register(app)

    while app:
        app.flush()

    app.fire(read(b"1\n2\r\n3\n4"))

    while app:
        app.flush()

    assert app.lines[0] == b"1"
    assert app.lines[1] == b"2"
    assert app.lines[2] == b"3"


def test_server():
    app = AppServer()
    buffers = defaultdict(bytes)
    Line(
        getBuffer=buffers.__getitem__,
        updateBuffer=buffers.__setitem__
    ).register(app)

    while app:
        app.flush()

    app.fire(read(1, b"1\n2\r\n3\n4"))
    app.fire(read(2, b"1\n2\r\n3\n4"))

    while app:
        app.flush()

    assert app.lines[0] == (1, b"1")
    assert app.lines[1] == (1, b"2")
    assert app.lines[2] == (1, b"3")

    assert app.lines[3] == (2, b"1")
    assert app.lines[4] == (2, b"2")
    assert app.lines[5] == (2, b"3")



#!/usr/bin/env python


import pytest


from os.path import dirname


from circuits import Event, Loader, Manager


class test(Event):
    """test Event"""


def test_main():
    m = Manager()
    loader = Loader(paths=[dirname(__file__)]).register(m)

    m.start()

    loader.load("app")

    x = m.fire(test())

    assert pytest.wait_for(x, "result")

    s = x.value
    assert s == "Hello World!"

    m.stop()



#!/usr/bin/env python


from pytest import fixture, skip, PLATFORM


if PLATFORM == "win32":
    skip("Broken on Windows")


from circuits import Component, Event
from circuits.net.events import close
from circuits.node import Node, remote
from circuits.net.sockets import UDPServer


class App(Component):

    ready = False
    value = False
    disconnected = False

    def foo(self):
        return "Hello World!"

    def ready(self, *args):
        self.ready = True

    def disconnect(self, component):
        self.disconnected = True

    def remote_value_changed(self, value):
        self.value = True


@fixture()
def bind(request, manager, watcher):
    server = UDPServer(0).register(manager)
    assert watcher.wait("ready")

    host, port = server.host, server.port

    server.fire(close())
    assert watcher.wait("closed")

    server.unregister()
    assert watcher.wait("unregistered")

    return host, port


@fixture()
def app(request, manager, watcher, bind):
    app = App().register(manager)
    node = Node().register(app)
    watcher.wait("ready")

    child = (App() + Node(bind))
    child.start(process=True)

    node.add("child", *bind)
    watcher.wait("connected")

    def finalizer():
        child.stop()

    request.addfinalizer(finalizer)

    return app


def test_return_value(app, watcher):
    e = Event.create("foo")
    e.notify = True

    r = remote(e, "child")
    r.notify = True

    value = app.fire(r)
    assert watcher.wait("remote_value_changed")

    assert value.value == "Hello World!"



#!/usr/bin/env python

try:
    from httplib import HTTPConnection
except ImportError:
    from http.client import HTTPConnection  # NOQA

from circuits.web import Controller


class Root(Controller):
    def index(self):
        return "Hello World!"


def test(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.auto_open = False
    connection.connect()

    for i in range(2):
        connection.request("GET", "/")
        response = connection.getresponse()
        assert response.status == 200
        assert response.reason == "OK"
        s = response.read()
        assert s == b"Hello World!"

    connection.close()



# Module:   conftest
# Date:     6th December 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""py.test config"""

import pytest

import sys
import threading
from time import sleep
from collections import deque

from circuits.core.manager import TIMEOUT
from circuits import handler, BaseComponent, Debugger, Manager


class Watcher(BaseComponent):

    def init(self):
        self._lock = threading.Lock()
        self.events = deque()

    @handler(channel="*", priority=999.9)
    def _on_event(self, event, *args, **kwargs):
        with self._lock:
            self.events.append(event)

    def wait(self, name, channel=None, timeout=6.0):
        for i in range(int(timeout / TIMEOUT)):
            if channel is None:
                with self._lock:
                    for event in self.events:
                        if event.name == name:
                            return True
            else:
                with self._lock:
                    for event in self.events:
                        if event.name == name and \
                                channel in event.channels:
                            return True

            sleep(TIMEOUT)


class Flag(object):
    status = False


class WaitEvent(object):

    def __init__(self, manager, name, channel=None, timeout=6.0):
        if channel is None:
            channel = getattr(manager, "channel", None)

        self.timeout = timeout
        self.manager = manager

        flag = Flag()

        @handler(name, channel=channel)
        def on_event(self, *args, **kwargs):
            flag.status = True

        self.handler = self.manager.addHandler(on_event)
        self.flag = flag

    def wait(self):
        try:
            for i in range(int(self.timeout / TIMEOUT)):
                if self.flag.status:
                    return True
                sleep(TIMEOUT)
        finally:
            self.manager.removeHandler(self.handler)


@pytest.fixture(scope="session")
def manager(request):
    manager = Manager()

    def finalizer():
        manager.stop()

    request.addfinalizer(finalizer)

    waiter = WaitEvent(manager, "started")
    manager.start()
    assert waiter.wait()

    if request.config.option.verbose:
        Debugger().register(manager)

    return manager


@pytest.fixture
def watcher(request, manager):
    watcher = Watcher().register(manager)

    def finalizer():
        waiter = WaitEvent(manager, "unregistered")
        watcher.unregister()
        waiter.wait()

    request.addfinalizer(finalizer)

    return watcher


def pytest_namespace():
    return dict((
        ("WaitEvent", WaitEvent),
        ("PLATFORM", sys.platform),
        ("PYVER", sys.version_info[:3]),
    ))



#!/usr/bin/env python

import pytest

from circuits import Component
from circuits.web import Controller
from circuits.web.client import parse_url
from circuits.net.sockets import TCPClient
from circuits.net.events import connect, write


class Client(Component):
    def __init__(self, *args, **kwargs):
        super(Client, self).__init__(*args, **kwargs)
        self._buffer = []
        self.done = False

    def read(self, data):
        self._buffer.append(data)
        if data.find(b"\r\n") != -1:
            self.done = True

    def buffer(self):
        return b''.join(self._buffer)


class Root(Controller):

    def index(self):
        return "Hello World!"


def test(webapp):
    transport = TCPClient()
    client = Client()
    client += transport
    client.start()

    host, port, resource, secure = parse_url(webapp.server.http.base)
    client.fire(connect(host, port))
    assert pytest.wait_for(transport, "connected")

    client.fire(write(b"GET / HTTP/1.1\r\n"))
    client.fire(write(b"Content-Type: text/plain\r\n\r\n"))
    assert pytest.wait_for(client, "done")

    client.stop()

    s = client.buffer().decode('utf-8').split('\r\n')[0]
    assert s == "HTTP/1.1 200 OK"



#!/usr/bin/env python


"""circuits Signal Handling

A modified version of the circuits Hello World example to demonstrate
how to customize signal handling and cause a delayed system terminate.
"""


from __future__ import print_function


from circuits import Component, Debugger, Event, Timer


class hello(Event):
    """hello Event"""


class App(Component):

    def hello(self):
        """Hello Event Handler"""

        print("Hello World!")

    def signal(self, event, signo, stack):
        Timer(3, Event.create("terminate")).register(self)
        print("Terminating in 3 seconds...")

    def terminate(self):
        self.stop()

    def started(self, component):
        """Started Event Handler

        This is fired internally when your application starts up
        and can be used to trigger events that only occur once
        during startup.
        """

        self.fire(hello())  # Fire hello Event


(App() + Debugger()).run()



# Module:   test_utils
# Date:     ...
# Author:   ...

"""test_utils

...
"""


from circuits import Event


class test(Event):
    """test Event"""


def test_events():
    from circuits.node.utils import dump_event, load_event

    e = test(1, 2, 3, foo="bar")
    e.success = True
    e.failure = False

    id = 1
    s = dump_event(e, id)
    x, id = load_event(s)

    assert hasattr(x, "args")
    assert hasattr(x, "kwargs")
    assert hasattr(x, "success")
    assert hasattr(x, "failure")
    assert hasattr(x, "channels")
    assert hasattr(x, "notify")


def test_values():
    from circuits.node.utils import dump_value, load_value
    from circuits.core.values import Value

    v = Value()
    v.value = 'foo'
    v.errors = False
    v.node_trn = 1

    s = dump_value(v)
    x, id, errors = load_value(s)

    assert v.value == x
    assert id == 1
    assert not errors



#!/usr/bin/env python

import pytest

import select
from socket import error as SocketError
from socket import EAI_NODATA, EAI_NONAME
from socket import socket, AF_INET, AF_INET6, SOCK_STREAM, has_ipv6

from circuits import Manager
from circuits.net.events import close, connect, write
from circuits.core.pollers import Select, Poll, EPoll, KQueue
from circuits.net.sockets import TCPServer, TCP6Server, TCPClient, TCP6Client

from .client import Client
from .server import Server


def wait_host(server):
    def checker(obj, attr):
        return all((getattr(obj, a) for a in attr))
    assert pytest.wait_for(server, ("host", "port"), checker)


def _pytest_generate_tests(metafunc, ipv6):
    metafunc.addcall(funcargs={"Poller": Select, "ipv6": ipv6})

    if hasattr(select, "poll"):
        metafunc.addcall(funcargs={"Poller": Poll, "ipv6": ipv6})

    if hasattr(select, "epoll"):
        metafunc.addcall(funcargs={"Poller": EPoll, "ipv6": ipv6})

    if hasattr(select, "kqueue"):
        metafunc.addcall(funcargs={"Poller": KQueue, "ipv6": ipv6})


def pytest_generate_tests(metafunc):
    _pytest_generate_tests(metafunc, ipv6=False)
    if has_ipv6:
        _pytest_generate_tests(metafunc, ipv6=True)


def test_tcp_basic(Poller, ipv6):
    m = Manager() + Poller()

    if ipv6:
        tcp_server = TCP6Server(("::1", 0))
        tcp_client = TCP6Client()
    else:
        tcp_server = TCPServer(0)
        tcp_client = TCPClient()
    server = Server() + tcp_server
    client = Client() + tcp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, "ready")
        assert pytest.wait_for(server, "ready")
        wait_host(server)

        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, "connected")
        assert pytest.wait_for(server, "connected")
        assert pytest.wait_for(client, "data", b"Ready")

        client.fire(write(b"foo"))
        assert pytest.wait_for(server, "data", b"foo")
        assert pytest.wait_for(client, "data", b"foo")

        client.fire(close())
        assert pytest.wait_for(client, "disconnected")
        assert pytest.wait_for(server, "disconnected")

        server.fire(close())
        assert pytest.wait_for(server, "closed")
    finally:
        m.stop()


def test_tcp_reconnect(Poller, ipv6):
    ### XXX: Apparently this doesn't work on Windows either?
    ### XXX: UPDATE: Apparently Broken on Windows + Python 3.2
    ### TODO: Need to look into this. Find out why...

    if pytest.PLATFORM == "win32" and pytest.PYVER[:2] >= (3, 2):
        pytest.skip("Broken on Windows on Python 3.2")

    m = Manager() + Poller()

    if ipv6:
        tcp_server = TCP6Server(("::1", 0))
        tcp_client = TCP6Client()
    else:
        tcp_server = TCPServer(0)
        tcp_client = TCPClient()
    server = Server() + tcp_server
    client = Client() + tcp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, "ready")
        assert pytest.wait_for(server, "ready")
        wait_host(server)

        # 1st connect
        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, "connected")
        assert pytest.wait_for(server, "connected")
        assert pytest.wait_for(client, "data", b"Ready")

        client.fire(write(b"foo"))
        assert pytest.wait_for(server, "data", b"foo")

        # disconnect
        client.fire(close())
        assert pytest.wait_for(client, "disconnected")

        # 2nd reconnect
        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, "connected")
        assert pytest.wait_for(server, "connected")
        assert pytest.wait_for(client, "data", b"Ready")

        client.fire(write(b"foo"))
        assert pytest.wait_for(server, "data", b"foo")

        client.fire(close())
        assert pytest.wait_for(client, "disconnected")
        assert pytest.wait_for(server, "disconnected")

        server.fire(close())
        assert pytest.wait_for(server, "closed")
    finally:
        m.stop()


def test_tcp_connect_closed_port(Poller, ipv6):
    ### FIXME: This test is wrong.
    ### We need to figure out the sequence of events on Windows
    ### for this scenario. I think if you attempt to connect to
    ### a shutdown listening socket (tcp server) you should get
    ### an error event as response.

    if pytest.PLATFORM == "win32":
        pytest.skip("Broken on Windows")

    m = Manager() + Poller()
    if ipv6:
        tcp_server = TCP6Server(("::1", 0))
        tcp_client = TCP6Client()
    else:
        tcp_server = TCPServer(0)
        tcp_client = TCPClient()
    server = Server() + tcp_server
    client = Client() + tcp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, "ready")
        assert pytest.wait_for(server, "ready")
        wait_host(server)

        host, port = server.host, server.port
        tcp_server._sock.close()

        # 1st connect
        client.fire(connect(host, port))
        assert pytest.wait_for(client, "connected")
        assert isinstance(client.error, SocketError)

        client.fire(write(b"foo"))
        assert pytest.wait_for(client, "disconnected")

        client.disconnected = False
        client.fire(write(b"foo"))
        assert pytest.wait_for(client, "disconnected", timeout=1.0) is None
    finally:
        m.stop()


def test_tcp_bind(Poller, ipv6):
    m = Manager() + Poller()

    if ipv6:
        sock = socket(AF_INET6, SOCK_STREAM)
        sock.bind(("::1", 0))
        sock.listen(5)
        _, bind_port, _, _ = sock.getsockname()
        sock.close()
        server = Server() + TCP6Server(("::1", 0))
        client = Client() + TCP6Client()
    else:
        sock = socket(AF_INET, SOCK_STREAM)
        sock.bind(("", 0))
        sock.listen(5)
        _, bind_port = sock.getsockname()
        sock.close()
        server = Server() + TCPServer(0)
        client = Client() + TCPClient()

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, "ready")
        assert pytest.wait_for(server, "ready")
        wait_host(server)

        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, "connected")
        assert pytest.wait_for(server, "connected")
        assert pytest.wait_for(client, "data", b"Ready")

        #assert server.client[1] == bind_port

        client.fire(write(b"foo"))
        assert pytest.wait_for(server, "data", b"foo")

        client.fire(close())
        assert pytest.wait_for(client, "disconnected")
        assert pytest.wait_for(server, "disconnected")

        server.fire(close())
        assert pytest.wait_for(server, "closed")
    finally:
        m.stop()


def test_tcp_lookup_failure(Poller, ipv6):
    m = Manager() + Poller()

    if ipv6:
        tcp_client = TCP6Client()
    else:
        tcp_client = TCPClient()
    client = Client() + tcp_client

    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, "ready")

        client.fire(connect("foo", 1234))
        assert pytest.wait_for(client, "error", lambda obj, attr: isinstance(getattr(obj, attr), SocketError))
        if pytest.PLATFORM == "win32":
            assert client.error.errno == 11004
        else:
            assert client.error.errno in (EAI_NODATA, EAI_NONAME,)
    finally:
        m.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from circuits.web import expose, Controller

from .helpers import urlopen


class Root(Controller):
    def __init__(self, *args, **kwargs):
        super(Root, self).__init__(*args, **kwargs)
        self += Hello()
        self += World()

    def index(self):
        return "index"

    def hello1(self):
        return "hello1"

    @expose("hello2")
    def hello2(self):
        return "hello2"

    def query(req, test):
        return 'query %s' % test


class Hello(Controller):
    channel = "/hello"

    def index(self):
        return 'hello index'

    def test(self):
        return 'hello test'

    def query(req, test):
        return 'hello query %s' % test


class World(Controller):
    channel = "/world"

    def index(self):
        return 'world index'

    def test(self):
        return 'world test'


def test_simple(webapp):
    url = "%s/hello1" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"hello1"


def test_expose(webapp):
    url = "%s/hello2" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"hello2"


def test_index(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b"index"


def test_controller_index(webapp):
    url = "%s/hello/" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"hello index"

    url = "%s/world/" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"world index"


def test_controller_expose(webapp):
    url = "%s/hello/test" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"hello test"

    url = "%s/world/test" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"world test"


def test_query(webapp):
    url = "%s/query?test=1" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"query 1"

    url = "%s/hello/query?test=2" % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b"hello query 2"



#!/usr/bin/env python

from circuits.web import Server
from circuits.web.controllers import expose, BaseController


class Root(BaseController):

    @expose("index")
    def index(self):
        """Index Request Handler

        BaseController(s) do not expose methods as request handlers.
        Request Handlers _must_ be exposed explicitly using the ``@expose``
        decorator.
        """

        return "Hello World!"

app = Server(("0.0.0.0", 8000))
Root().register(app)
app.run()



#!/usr/bin/env python

from circuits import handler, Event, BaseComponent


class test(Event):
    """test Event"""


class App(BaseComponent):

    @handler("test")
    def _on_test(self, event):
        try:
            return "Hello World!"
        finally:
            event.stop()

    def _on_test2(self):
        pass  # Never reached


def test_main():
    app = App()
    while app:
        app.flush()
    x = app.fire(test())
    app.flush()
    assert x.value == "Hello World!"



#!/usr/bin/env python

import pytest
if pytest.PYVER[:2] == (3, 3):
    pytest.skip("Broken on Python 3.3")

from circuits.web import Server
from circuits.web.wsgi import Gateway

from .helpers import urlopen


def hello(environ, start_response):
    status = "200 OK"
    response_headers = [("Content-type", "text/plain")]
    start_response(status, response_headers)
    return "Hello World!"


def foobar(environ, start_response):
    status = "200 OK"
    response_headers = [("Content-type", "text/plain")]
    start_response(status, response_headers)
    return "FooBar!"


@pytest.fixture
def apps(request):
    return {
        "/": hello,
        "/foobar": foobar
    }


def test(apps):
    server = Server(0)
    Gateway(apps).register(server)

    waiter = pytest.WaitEvent(server, "ready")
    server.start()
    waiter.wait()

    f = urlopen(server.http.base)
    s = f.read()
    assert s == b"Hello World!"

    f = urlopen("{0:s}/foobar/".format(server.http.base))
    s = f.read()
    assert s == b"FooBar!"

    server.stop()



# Module:   message
# Date:     11th August 2014
# Author:   James Mills <prologic@shortcircuit.net.au>


"""Internet Relay Chat message"""


from .utils import parsemsg


class Error(Exception):
    """Error Exception"""


class Message(object):

    def __init__(self, command, *args, **kwargs):
        for arg in args[:-1]:
            if " " in arg:
                raise Error("Space can only appear in the very last arg")

        self.command = command
        self.args = list(filter(lambda x: x is not None, list(args)))
        self.prefix = str(kwargs["prefix"]) if "prefix" in kwargs else None

        self.encoding = kwargs.get("encoding", "utf-8")
        self.add_nick = kwargs.get("add_nick", False)

    @staticmethod
    def from_string(s):
        if len(s) > 512:
            raise Error("Message must not be longer than 512 characters")

        prefix, command, args = parsemsg(s)

        return Message(command, *args, prefix=prefix)

    def __bytes__(self):
        return str(self).encode(self.encoding)

    def __str__(self):
        args = self.args[:]
        for arg in args[:-1]:
            if arg is not None and " " in arg:
                raise Error("Space can only appear in the very last arg")

        if len(args) > 0 and " " in args[-1]:
            args[-1] = ":{0:s}".format(args[-1])

        return "{prefix:s}{command:s} {args:s}\r\n".format(
            prefix=(
                ":{0:s} ".format(self.prefix)
                if self.prefix is not None
                else ""
            ),
            command=str(self.command),
            args=" ".join(args)
        )

    def __repr__(self):
        return "\"{0:s}\"".format(str(self)[:-2])

    def __eq__(self, other):
        return isinstance(other, Message) \
            and self.prefix == other.prefix \
            and self.command == other.command \
            and self.args == other.args



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Main

circutis.web Web Server and Testing Tool.
"""


import os
from sys import stderr
from hashlib import md5
from optparse import OptionParser
from wsgiref.validate import validator
from wsgiref.simple_server import make_server

try:
    import hotshot
    import hotshot.stats
except ImportError:
    hostshot = None


import circuits
from circuits import handler, Component, Manager, Debugger

from circuits.core.pollers import Select
from circuits.tools import inspect, graph

from circuits.web.wsgi import Application
from circuits.web.tools import check_auth, digest_auth
from circuits.web import BaseServer, Controller, Logger, Server, Static

try:
    from circuits.core.pollers import Poll
except ImportError:
    Poll = None  # NOQA

try:
    from circuits.core.pollers import EPoll
except ImportError:
    EPoll = None  # NOQA


USAGE = "%prog [options] [docroot]"
VERSION = "%prog v" + circuits.__version__


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        "-b", "--bind",
        action="store", type="string", default="0.0.0.0:8000", dest="bind",
        help="Bind to address:[port]"
    )

    parser.add_option(
        "-l", "--logging",
        action="store_true", default=False, dest="logging",
        help="Enable logging of requests"
    )

    parser.add_option(
        "-p", "--passwd",
        action="store", default=None, dest="passwd",
        help="Location to passwd file for Digest Auth"
    )

    parser.add_option(
        "-j", "--jobs",
        action="store", type="int", default=0, dest="jobs",
        help="Specify no. of jobs/processes to start"
    )

    parser.add_option(
        "", "--poller",
        action="store", type="string", default="select", dest="poller",
        help="Specify type of poller to use"
    )

    parser.add_option(
        "", "--server",
        action="store", type="string", default="server", dest="server",
        help="Specify server to use"
    )

    parser.add_option(
        "", "--profile",
        action="store_true", default=False, dest="profile",
        help="Enable execution profiling support"
    )

    parser.add_option(
        "", "--debug",
        action="store_true", default=False, dest="debug",
        help="Enable debug mode"
    )

    parser.add_option(
        "", "--validate",
        action="store_true", default=False, dest="validate",
        help="Enable WSGI validation mode"
    )

    opts, args = parser.parse_args()

    return opts, args


class Authentication(Component):

    channel = "web"

    realm = "Secure Area"
    users = {"admin": md5("admin").hexdigest()}

    def __init__(self, channel=channel, realm=None, passwd=None):
        super(Authentication, self).__init__(self, channel=channel)

        if realm is not None:
            self.realm = realm

        if passwd is not None:
            with open(passwd, "r") as f:
                lines = (line.strip() for line in f)
                self.users = dict((line.split(":", 1) for line in lines))

    @handler("request", priority=10)
    def request(self, event, request, response):
        if not check_auth(request, response, self.realm, self.users):
            event.stop()
            return digest_auth(request, response, self.realm, self.users)


class HelloWorld(Component):

    channel = "web"

    def request(self, request, response):
        return "Hello World!"


class Root(Controller):

    def hello(self):
        return "Hello World!"


def select_poller(poller):
    if poller == "poll":
        if Poll is None:
            stderr.write(
                "No poll support available - defaulting to Select..."
            )
            Poller = Select
        else:
            Poller = Poll
    elif poller == "epoll":
        if EPoll is None:
            stderr.write(
                "No epoll support available - defaulting to Select..."
            )
            Poller = Select
        else:
            Poller = EPoll
    else:
        Poller = Select

    return Poller


def parse_bind(bind):
    if ":" in bind:
        address, port = bind.split(":")
        port = int(port)
    else:
        address, port = bind, 8000

    return (address, port)


def main():
    opts, args = parse_options()

    bind = parse_bind(opts.bind)

    if opts.validate:
        application = (Application() + Root())
        app = validator(application)

        httpd = make_server(bind[0], bind[1], app)
        httpd.serve_forever()

        raise SystemExit(0)

    manager = Manager()

    opts.debug and Debugger().register(manager)

    Poller = select_poller(opts.poller.lower())
    Poller().register(manager)

    if opts.server.lower() == "base":
        BaseServer(bind).register(manager)
        HelloWorld().register(manager)
    else:
        Server(bind).register(manager)
        Root().register(manager)

    docroot = os.getcwd() if not args else args[0]

    Static(docroot=docroot, dirlisting=True).register(manager)

    opts.passwd and Authentication(passwd=opts.passwd).register(manager)

    opts.logging and Logger().register(manager)

    if opts.profile and hotshot:
        profiler = hotshot.Profile(".profile")
        profiler.start()

    if opts.debug:
        print(graph(manager, name="circuits.web"))
        print()
        print(inspect(manager))

    for i in range(opts.jobs):
        manager.start(process=True)

    manager.run()

    if opts.profile and hotshot:
        profiler.stop()
        profiler.close()

        stats = hotshot.stats.load(".profile")
        stats.strip_dirs()
        stats.sort_stats("time", "calls")
        stats.print_stats(20)


if __name__ == "__main__":
    main()



# Module:   dispatcher
# Date:     13th September 2007
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Dispatcher

This module implements a basic URL to Channel dispatcher.
This is the default dispatcher used by circuits.web
"""

try:
    from urllib import quote, unquote
except ImportError:
    from urllib.parse import quote, unquote  # NOQA

from circuits.six import text_type

from circuits import handler, BaseComponent, Event

from circuits.web.utils import parse_qs
from circuits.web.events import response
from circuits.web.errors import httperror
from circuits.web.processors import process
from circuits.web.controllers import BaseController


def resolve_path(paths, parts):

    def rebuild_path(url_parts):
        return '/%s' % '/'.join(url_parts)

    left_over = []

    while parts:
        if rebuild_path(parts) in paths:
            yield rebuild_path(parts), left_over
        left_over.insert(0, parts.pop())

    if '/' in paths:
        yield '/', left_over


def resolve_methods(parts):
    if parts:
        method = parts[0]
        vpath = parts[1:]
        yield method, vpath

    yield 'index', parts


def find_handlers(req, paths):
    def get_handlers(path, method):
        component = paths[path]
        return component._handlers.get(method, None)

    def accepts_vpath(handlers, vpath):
        args_no = len(vpath)
        return all(
            len(h.args) == args_no or h.varargs or (
                h.defaults is not None and args_no <= len(h.defaults)
            )
            for h in handlers
        )

    # Split /hello/world to ['hello', 'world']
    starting_parts = [x for x in req.path.strip("/").split("/") if x]

    for path, parts in resolve_path(paths, starting_parts):
        handlers = get_handlers(path, req.method)
        if handlers:
            return handlers, req.method, path, parts

        for method, vpath in resolve_methods(parts):
            handlers = get_handlers(path, method)
            if handlers and (not vpath or accepts_vpath(handlers, vpath)):
                req.index = (method == 'index')
                return handlers, method, path, vpath
            else:
                method, vpath = "index", [method] + vpath
                handlers = get_handlers(path, method)
                if handlers and (not vpath or accepts_vpath(handlers, vpath)):
                    req.index = True
                    return handlers, method, path, vpath

    return [], None, None, None


class Dispatcher(BaseComponent):

    channel = "web"

    def __init__(self, **kwargs):
        super(Dispatcher, self).__init__(**kwargs)

        self.paths = dict()

    @handler("registered", channel="*")
    def _on_registered(self, component, manager):
        if (isinstance(component, BaseController) and component.channel not
                in self.paths):
            self.paths[component.channel] = component

    @handler("unregistered", channel="*")
    def _on_unregistered(self, component, manager):
        if (isinstance(component, BaseController) and component.channel in
                self.paths):
            del self.paths[component.channel]

    @handler("request", priority=0.1)
    def _on_request(self, event, req, res, peer_cert=None):
        if peer_cert:
            event.peer_cert = peer_cert

        handlers, name, channel, vpath = find_handlers(req, self.paths)

        if name is not None and channel is not None:
            event.kwargs = parse_qs(req.qs)
            process(req, event.kwargs)

            if vpath:
                event.args += tuple(vpath)

            if isinstance(name, text_type):
                name = str(name)

            return self.fire(
                Event.create(
                    name, *event.args, **event.kwargs
                ),
                channel
            )

    @handler("request_value_changed")
    def _on_request_value_changed(self, value):
        if value.handled:
            return

        req, res = value.event.args[:2]
        if value.result and not value.errors:
            res.body = value.value
            self.fire(response(res))
        elif value.promise:
            value.event.notify = True
        else:
            # Errors are handled by the ``HTTP`` Protocol Component
            return



# Module:   server
# Date:     6th November 2008
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Web Servers

This module implements the several Web Server components.
"""


from sys import stderr


from circuits import io
from circuits.net.events import close, read, write
from circuits.net.sockets import TCPServer, UNIXServer
from circuits.core import handler, BaseComponent, Timer

from .http import HTTP
from .events import terminate
from .dispatchers import Dispatcher


class BaseServer(BaseComponent):
    """Create a Base Web Server

    Create a Base Web Server (HTTP) bound to the IP Address / Port or
    UNIX Socket specified by the 'bind' parameter.

    :ivar server: Reference to underlying Server Component

    :param bind: IP Address / Port or UNIX Socket to bind to.
    :type bind: Instance of int, list, tuple or str

    The 'bind' parameter is quite flexible with what valid values it accepts.

    If an int is passed, a TCPServer will be created. The Server will be
    bound to the Port given by the 'bind' argument and the bound interface
    will default (normally to  "0.0.0.0").

    If a list or tuple is passed, a TCPServer will be created. The
    Server will be bound to the Port given by the 2nd item in the 'bind'
    argument and the bound interface will be the 1st item.

    If a str is passed and it contains the ':' character, this is
    assumed to be a request to bind to an IP Address / Port. A TCpServer
    will thus be created and the IP Address and Port will be determined
    by splitting the string given by the 'bind' argument.

    Otherwise if a str is passed and it does not contain the ':'
    character, a file path is assumed and a UNIXServer is created and
    bound to the file given by the 'bind' argument.
    """

    channel = "web"

    def __init__(self, bind, encoding="utf-8", secure=False, certfile=None,
                 channel=channel):
        "x.__init__(...) initializes x; see x.__class__.__doc__ for signature"

        super(BaseServer, self).__init__(channel=channel)

        if isinstance(bind, (int, list, tuple,)):
            SocketType = TCPServer
        else:
            SocketType = TCPServer if ":" in bind else UNIXServer

        self.server = SocketType(
            bind,
            secure=secure,
            certfile=certfile,
            channel=channel
        ).register(self)

        self.http = HTTP(
            self, encoding=encoding, channel=channel
        ).register(self)

    @property
    def host(self):
        if hasattr(self, "server"):
            return self.server.host

    @property
    def port(self):
        if hasattr(self, "server"):
            return self.server.port

    @property
    def secure(self):
        if hasattr(self, "server"):
            return self.server.secure

    @handler("connect")
    def _on_connect(self, *args, **kwargs):
        """Dummy Event Handler for connect"""

    @handler("closed")
    def _on_closed(self, *args, **kwargs):
        """Dummy Event Handler for closed"""

    @handler("signal")
    def _on_signal(self, *args, **kwargs):
        """signal Event Handler"""

        self.fire(close())
        Timer(3, terminate()).register(self)

    @handler("terminate")
    def _on_terminate(self):
        raise SystemExit(0)

    @handler("ready")
    def _on_ready(self, server, bind):
        stderr.write(
            "{0:s} ready! Listening on: {1:s}\n".format(
                self.http.version, self.http.base
            )
        )


class Server(BaseServer):
    """Create a Web Server

    Create a Web Server (HTTP) complete with the default Dispatcher to
    parse requests and posted form data dispatching to appropriate
    Controller(s).

    See: circuits.web.servers.BaseServer
    """

    def __init__(self, bind, **kwargs):
        "x.__init__(...) initializes x; see x.__class__.__doc__ for signature"

        super(Server, self).__init__(bind, **kwargs)

        Dispatcher(channel=self.channel).register(self.http)


class FakeSock():
    def getpeername(self):
        return (None, None)


class StdinServer(BaseComponent):

    channel = "web"

    def __init__(self, encoding="utf-8", channel=channel):
        super(StdinServer, self).__init__(channel=channel)

        self.server = (io.stdin + io.stdout).register(self)
        self.http = HTTP(
            self, encoding=encoding, channel=channel
        ).register(self)

        Dispatcher(channel=self.channel).register(self)

    @property
    def host(self):
        return io.stdin.filename

    @property
    def port(self):
        return 0

    @property
    def secure(self):
        return False

    @handler("read", channel="stdin")
    def read(self, data):
        self.fire(read(FakeSock(), data))

    @handler("write")
    def write(self, sock, data):
        self.fire(write(data))



# Module:   wrappers
# Date:     13th September 2007
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Request/Response Wrappers

This module implements the Request and Response objects.
"""


from time import time
from io import BytesIO
from functools import partial

try:
    from Cookie import SimpleCookie
except ImportError:
    from http.cookies import SimpleCookie  # NOQA

from .url import parse_url
from .headers import Headers
from ..six import binary_type
from .errors import httperror
from circuits.net.sockets import BUFSIZE
from .constants import HTTP_STATUS_CODES, SERVER_VERSION

try:
    unicode
except NameError:
    unicode = str

try:
    from email.utils import formatdate
    formatdate = partial(formatdate, usegmt=True)
except ImportError:
    from rfc822 import formatdate as HTTPDate  # NOQA


def file_generator(input, chunkSize=BUFSIZE):
    chunk = input.read(chunkSize)
    while chunk:
        yield chunk
        chunk = input.read(chunkSize)
    input.close()


class Host(object):
    """An internet address.

    name should be the client's host name. If not available (because no DNS
    lookup is performed), the IP address should be used instead.
    """

    ip = "0.0.0.0"
    port = 80
    name = "unknown.tld"

    def __init__(self, ip, port, name=None):
        self.ip = ip
        self.port = port
        if name is None:
            name = ip
        self.name = name

    def __repr__(self):
        return "Host(%r, %r, %r)" % (self.ip, self.port, self.name)


class HTTPStatus(object):

    __slots__ = ("_reason", "_status",)

    def __init__(self, status=200, reason=None):
        self._status = status
        self._reason = reason or HTTP_STATUS_CODES.get(status, "")

    def __int__(self):
        return self._status

    def __lt__(self, other):
        if isinstance(other, int):
            return self._status < other
        return super(HTTPStatus, self).__lt__(other)

    def __gt__(self, other):
        if isinstance(other, int):
            return self._status > other
        return super(HTTPStatus, self).__gt__(other)

    def __le__(self, other):
        if isinstance(other, int):
            return self._status <= other
        return super(HTTPStatus, self).__le__(other)

    def __ge__(self, other):
        if isinstance(other, int):
            return self._status >= other
        return super(HTTPStatus, self).__ge__(other)

    def __eq__(self, other):
        if isinstance(other, int):
            return self._status == other
        return super(HTTPStatus, self).__eq__(other)

    def __str__(self):
        return "{0:d} {1:s}".format(self._status, self._reason)

    def __repr__(self):
        return "<Status (status={0:d} reason={1:s}>".format(
            self._status, self._reason
        )

    def __format__(self, format_spec):
        return format(str(self), format_spec)

    @property
    def status(self):
        return self._status

    @property
    def reason(self):
        return self._reason


class Request(object):
    """Creates a new Request object to hold information about a request.

    :param sock: The socket object of the request.
    :type  sock: socket.socket

    :param method: The requested method.
    :type  method: str

    :param scheme: The requested scheme.
    :type  scheme: str

    :param path: The requested path.
    :type  path: str

    :param protocol: The requested protocol.
    :type  protocol: str

    :param qs: The query string of the request.
    :type  qs: str
    """

    server = None
    """:cvar: A reference to the underlying server"""

    scheme = "http"
    protocol = (1, 1)
    host = ""
    local = Host("127.0.0.1", 80)
    remote = Host("", 0)

    index = None
    script_name = ""

    login = None
    handled = False

    def __init__(self, sock, method="GET", scheme="http", path="/",
                 protocol=(1, 1), qs="", headers=None, server=None):
        "initializes x; see x.__class__.__doc__ for signature"

        self.sock = sock
        self.method = method
        self.scheme = scheme or Request.scheme
        self.path = path
        self.protocol = protocol
        self.qs = qs

        self.headers = headers or Headers()
        self.server = server

        self.cookie = SimpleCookie()

        if sock is not None:
            name = sock.getpeername()
            if name is not None:
                self.remote = Host(*name)
            else:
                name = sock.getsockname()
                self.remote = Host(name, "", name)

        cookie = self.headers.get("Cookie")
        if cookie is not None:
            self.cookie.load(cookie)

        self.body = BytesIO()

        if self.server is not None:
            self.local = Host(self.server.host, self.server.port)

        try:
            host = self.headers["Host"]
            if ":" in host:
                parts = host.split(":", 1)
                host = parts[0]
                port = int(parts[1])
            else:
                port = 443 if self.scheme == "https" else 80
        except KeyError:
            host = self.local.name or self.local.ip
            port = getattr(self.server, "port")

        self.host = host
        self.port = port

        base = "{0:s}://{1:s}{2:s}/".format(
            self.scheme,
            self.host,
            ":{0:d}".format(self.port) if self.port not in (80, 443) else ""
        )
        self.base = parse_url(base)

        url = "{0:s}{1:s}{2:s}".format(
            base,
            self.path,
            "?{0:s}".format(self.qs) if self.qs else ""
        )
        self.uri = parse_url(url)
        self.uri.sanitize()

    def __repr__(self):
        protocol = "HTTP/%d.%d" % self.protocol
        return "<Request %s %s %s>" % (self.method, self.path, protocol)


class Body(object):
    """Response Body"""

    def __get__(self, response, cls=None):
        if response is None:
            return self
        else:
            return response._body

    def __set__(self, response, value):
        if response == value:
            return

        if isinstance(value, binary_type):
            if value:
                value = [value]
            else:
                value = []
        elif hasattr(value, "read"):
            response.stream = True
            value = file_generator(value)
        elif isinstance(value, httperror):
            value = [str(value)]
        elif value is None:
            value = []

        response._body = value


class Status(object):
    """Response Status"""

    def __get__(self, response, cls=None):
        if response is None:
            return self
        else:
            return response._status

    def __set__(self, response, value):
        value = HTTPStatus(value) if isinstance(value, int) else value

        response._status = value


class Response(object):
    """Response(sock, request) -> new Response object

    A Response object that holds the response to
    send back to the client. This ensure that the correct data
    is sent in the correct order.
    """

    body = Body()
    status = Status()

    done = False
    close = False
    stream = False
    chunked = False

    def __init__(self, request, encoding='utf-8', status=None):
        "initializes x; see x.__class__.__doc__ for signature"

        self.request = request
        self.encoding = encoding

        self._body = []
        self._status = HTTPStatus(status if status is not None else 200)

        self.time = time()

        self.headers = Headers()
        self.headers["Date"] = formatdate()

        if self.request.server is not None:
            self.headers.add_header("Server", self.request.server.http.version)
        else:
            self.headers.add_header("X-Powered-By", SERVER_VERSION)

        self.cookie = self.request.cookie

        self.protocol = "HTTP/%d.%d" % self.request.protocol

    def __repr__(self):
        return "<Response %s %s (%d)>" % (
            self.status,
            self.headers.get("Content-Type"),
            (len(self.body) if isinstance(self.body, str) else 0)
        )

    def __str__(self):
        self.prepare()
        protocol = self.protocol
        status = "{0:s}".format(self.status)
        return "{0:s} {1:s}\r\n".format(protocol, status)

    def __bytes__(self):
        return str(self).encode(self.encoding)

    def prepare(self):
        # Set a default content-Type if we don't have one.
        self.headers.setdefault(
            "Content-Type", "text/html; charset={0:s}".format(self.encoding)
        )

        cLength = None
        if self.body is not None:
            if isinstance(self.body, bytes):
                cLength = len(self.body)
            elif isinstance(self.body, unicode):
                cLength = len(self.body.encode(self.encoding))
            elif isinstance(self.body, list):
                cLength = sum(
                    [
                        len(s.encode(self.encoding))
                        if not isinstance(s, bytes)
                        else len(s) for s in self.body
                        if s is not None
                    ]
                )

        if cLength is not None:
            self.headers["Content-Length"] = str(cLength)

        for k, v in self.cookie.items():
            self.headers.add_header("Set-Cookie", v.OutputString())

        status = self.status

        if status == 413:
            self.close = True
        elif "Content-Length" not in self.headers:
            if status < 200 or status in (204, 205, 304):
                pass
            else:
                if self.protocol == "HTTP/1.1" \
                        and self.request.method != "HEAD" \
                        and self.request.server is not None \
                        and not cLength == 0:
                    self.chunked = True
                    self.headers.add_header("Transfer-Encoding", "chunked")
                else:
                    self.close = True

        if (self.request.server is not None
                and "Connection" not in self.headers):
            if self.protocol == "HTTP/1.1":
                if self.close:
                    self.headers.add_header("Connection", "close")
            else:
                if not self.close:
                    self.headers.add_header("Connection", "Keep-Alive")

        if self.headers.get("Transfer-Encoding", "") == "chunked":
            self.chunked = True



#!/usr/bin/env python

import pytest
if pytest.PLATFORM == "win32":
    pytest.skip("Unsupported Platform")

from io import BytesIO
from circuits.io import File
from circuits import Component
from circuits.io.events import write, close


class FileApp(Component):

    def init(self, *args, **kwargs):
        self.file = File(*args, **kwargs).register(self)

        self.eof = False
        self.closed = False
        self.buffer = BytesIO()

    def read(self, data):
        self.buffer.write(data)

    def eof(self):
        self.eof = True

    def closed(self):
        self.closed = True


def test_open_fileobj(manager, watcher, tmpdir):
    filename = str(tmpdir.ensure("helloworld.txt"))
    with open(filename, "w") as f:
        f.write("Hello World!")

    fileobj = open(filename, "r")

    app = FileApp(fileobj).register(manager)
    assert watcher.wait("opened", app.file.channel)

    assert watcher.wait("eof", app.file.channel)
    assert app.eof

    app.fire(close(), app.file.channel)
    assert watcher.wait("closed", app.file.channel)
    assert app.closed

    app.unregister()
    assert watcher.wait("unregistered")

    s = app.buffer.getvalue()
    assert s == b"Hello World!"


def test_read_write(manager, watcher, tmpdir):
    filename = str(tmpdir.ensure("helloworld.txt"))

    app = FileApp(filename, "w").register(manager)
    assert watcher.wait("opened", app.file.channel)

    app.fire(write(b"Hello World!"), app.file.channel)
    assert watcher.wait("write", app.file.channel)

    app.fire(close(), app.file.channel)
    assert watcher.wait("closed", app.file.channel)
    assert app.closed

    app.unregister()
    assert watcher.wait("unregistered")

    app = FileApp(filename, "r").register(manager)
    assert watcher.wait("opened", app.file.channel)

    assert watcher.wait("eof", app.file.channel)
    assert app.eof

    app.fire(close(), app.file.channel)
    assert watcher.wait("closed", app.file.channel)
    assert app.closed

    app.unregister()
    assert watcher.wait("unregistered")

    s = app.buffer.getvalue()
    assert s == b"Hello World!"
